// ------------------------------
// Filename      : cube.cpp
// Author        : Sven Maerivoet
// Last modified : 30/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "cube.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

using namespace std;

// ---------------
// class CubeGLApp
// ---------------

// -----------
// constructor
// -----------
CubeGLApp::CubeGLApp(void)
{
}

// ----------
// destructor
// ----------
CubeGLApp::~CubeGLApp(void)
{
  // destructors for RGBApixmaps are automatically called
  // when the container-vector goes out of scope
}

// -----------------------
// explicit initialisation
// -----------------------
void CubeGLApp::explicitInitialize(void)
{
  int i;

  fUseTICOTextures = false;

  fAngleIncrement = 5.0;
  fAngle = 0.0;

  fTICOTextures.resize(6);
  for (i = 0; i < 6; ++i) {
    fTICOTextures[i] = new RGBApixmap();
  }
  fEscherTexture = new RGBApixmap();

  if (!fTICOTextures[0]->loadFromBMPFile("data/textures/other/tico-logo.bmp")) {
    exit(0);
  }
  if (!fTICOTextures[1]->loadFromBMPFile("data/textures/other/tico-arickx.bmp")) {
    exit(0);
  }
  if (!fTICOTextures[2]->loadFromBMPFile("data/textures/other/tico-broeckhove.bmp")) {
    exit(0);
  }
  if (!fTICOTextures[3]->loadFromBMPFile("data/textures/other/tico-saenen.bmp")) {
    exit(0);
  }
  if (!fTICOTextures[4]->loadFromBMPFile("data/textures/other/tico-stuer.bmp")) {
    exit(0);
  }
  if (!fTICOTextures[5]->loadFromBMPFile("data/textures/other/tico-vanroose.bmp")) {
    exit(0);
  }

  for (i = 0; i < 6; ++i) {
    fTICOTextures[i]->setTextureID(getUniqueTextureID());
  }

  if (!fEscherTexture->loadFromBMPFile("data/textures/other/escher-fishes.bmp")) {
    exit(0);
  }
  fEscherTexture->setTextureID(getUniqueTextureID());

  switchLightingOn();

  setCameraSlideStepSize(5.0);
  setCameraRotateStepSize(5.0);

  setCamera(Point3D(0.0,0.0,50.0),
            Point3D(0.0,0.0,0.0),
            Vector3D(0.0,1.0,0.0));
  setCameraPerspectiveShape(45.0,getWorldWindowAspectRatio(),
                            1.0,10000.0);
}

// ------------------
// callback-functions
// ------------------
void CubeGLApp::displayFunc(void)
{
  setBackgroundColor(Color(0.2,0.0,0.3));
  clearScreen();

  switchLightOn(kLight1);
  setLightDirection(kLight1,Vector3D(getCameraEyePoint().fX,
                                     getCameraEyePoint().fY,
                                     getCameraEyePoint().fZ));

  disableGlowingTextures();

  double size = 10.0;
  size /= 2.0;

  Point3D p1(-size,+size,-size);
  Point3D p2(-size,-size,-size);
  Point3D p3(+size,-size,-size);
  Point3D p4(+size,+size,-size);

  Point3D q1(-size,+size,+size);
  Point3D q2(-size,-size,+size);
  Point3D q3(+size,-size,+size);
  Point3D q4(+size,+size,+size);

  pushCT();
    rotate3D(fAngle,1.0,1.0,1.0);

    if (fUseTICOTextures) {
      textureQuad(fTICOTextures[0],
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  q1,q2,q3,q4);
      textureQuad(fTICOTextures[1],
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  q4,q3,p3,p4);
      textureQuad(fTICOTextures[2],
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  p4,p3,p2,p1);
      textureQuad(fTICOTextures[3],
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  p1,p2,q2,q1);
      textureQuad(fTICOTextures[4],
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  p1,q1,q4,p4);
      textureQuad(fTICOTextures[5],
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  q2,p2,p3,q3);
    }
    else {
      textureQuad(fEscherTexture,
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  q1,q2,q3,q4);
      textureQuad(fEscherTexture,
                  Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),Point2D(0.0,1.0),
                  q4,q3,p3,p4);
      textureQuad(fEscherTexture,
                  Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                  p4,p3,p2,p1);
      textureQuad(fEscherTexture,
                  Point2D(1.0,1.0),Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),
                  p1,p2,q2,q1);
      textureQuad(fEscherTexture,
                  Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),Point2D(0.0,1.0),
                  p1,q1,q4,p4);
      textureQuad(fEscherTexture,
                  Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),Point2D(0.0,1.0),
                  q2,p2,p3,q3);
    }
  popCT();
}

void CubeGLApp::keyboardFunc(unsigned char key, int x, int y)
{
  TOpenGLApp::keyboardFunc(key,x,y);

  switch(key) {
    case 'r' : fUseTICOTextures = false;
               break;
    case 'R' : fUseTICOTextures = true;
               break;

    case 'g' : disableGoodTextureInterpolation();
               break;
    case 'G' : enableGoodTextureInterpolation();
               break;

    case '-' : if (fAngleIncrement > 0.0) {
                 fAngleIncrement -= 0.5;
               }
               break;
    case '+' : fAngleIncrement += 0.5;
               break;
  }

  if (fScheduleRedisplay) {
    redisplay();
  }
}

void CubeGLApp::idleFunc(void)
{
  TOpenGLApp::idleFunc();

  fAngle += fAngleIncrement;
  if (fAngle >= 360.0) {
    fAngle = 0.0;
  }

  redisplay();
}
