// ----------------------------------------------
// ALTERNATIVE FOR BoxExtent::rayIntersectsExtent
// ----------------------------------------------

bool BoxExtent::rayIntersectsExtent(const Ray& ray) const
{
  // setup bounding-box edge-points
  Vector3Dd a(fLeft,fBottom,fBack);
  Vector3Dd b(fRight,fTop,fFront);

  // check if the ray hits one of the six bounding faces
  if (rayIntersectsFace(Vector3D(0.0,0.0,1.0),
                        -b.fZ,ray,
                        a.fX,a.fY,b.fX,b.fY,0)) {
    return true;
  }
  else if (rayIntersectsFace(Vector3D(0.0,0.0,-1.0),
                             a.fZ,ray,
                             a.fX,a.fY,b.fX,b.fY,1)) {
    return true;
  }
  else if (rayIntersectsFace(Vector3Df(-1.0,0.0,0.0),
                             a.fX,ray,
                             a.fZ,a.fY,b.fZ,b.fY,2)) {
    return true;
  }
  else if (rayIntersectsFace(Vector3Df(1.0,0.0,0.0),
                             -b.fX,ray,
                             a.fZ,a.fY,b.fZ,b.fY,3)) {
    return true;
  }
  else if (rayIntersectsFace(Vector3Df(0.0,1.0,0.0),
                             -b.fY,ray,
                             a.fX,a.fZ,b.fX,b.fZ,4)) {
    return true;
  }
  else if (rayIntersectsFace(Vector3Df(0.0,-1.0,0.0),
                             a.fY,ray,
                             a.fX,a.fZ,b.fX,b.fZ,5)) {
    return true;
  }
  else {
    return false;
  }
}

bool BoxExtent::rayIntersectsFace(const Vector3Df& normal,
                                  double distanceToOrigin,
                                  const Ray& ray,
                                  double minA, double minB, double maxA, double maxB,
                                  int face) const
{
  double vd = ray.fDirection.inner(normal);

  if (vd > 0.0) {
    // ray is parallel to plane
    return false;
  }

  double v0 = -(Vector3Dd(ray.fStart).inner(normal) + distanceToOrigin);
  double t = v0 / vd;

  // check if the box lies behind the eye
  if (t < 0.0) {
    return false;
  }

  // the ray hits the plane
  // calculate intersection point and see if it lies inside the quadrilateral
  double px = ray.fStart.fX + (t * ray.fDirection.fX);
  double py = ray.fStart.fY + (t * ray.fDirection.fY);
  double pz = ray.fStart.fZ + (t * ray.fDirection.fZ);

  // now compare with the correct face
  switch (face) {
    case 2 :
    case 3 : px = pz;
             break;
    case 4 :
    case 5 : py = pz;
             break;
  }

  return ((px >= minA) &&
          (px <= maxA) &&
          (py >= minB) &&
          (py <= maxB));
}
