// --------------------------------------------
// Filename      : generate-fractal-spheres.cpp
// Author        : Sven Maerivoet
// Last modified : 08/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// --------------------------------------------

#include "opengl/primitives.h"
#include "opengl/rng.h"
#include <iostream>
#include <math.h>
#include <string>

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  stringstream strstream
#endif

using namespace std;

int kLeftMargin;

double getValue(string str)
{
  stringstream tmp;
  double number;

  tmp << str;
  if (!(tmp >> number)) {
    cerr << "getValue() -> error getting value [" << str << "]." << endl;
    exit(1);
  }
  else {
    char t;
    if (tmp >> t) {
      cerr << "getValue() -> bum chars in number [" << str << "]." << endl;
      exit(1);
    } 
  }

  return number;
}

void makeLeftMargin(void)
{
  for (int i = 0; i < kLeftMargin; ++i) {
    cout << " ";
  }
}

void generateOneSphere(double mx,
                       double my,
                       double mz,
                       double radius)
{
  makeLeftMargin(); cout << "push" << endl;
  makeLeftMargin(); cout << "  translate " << mx << " "
                                           << my << " "
                                           << mz << endl;
  makeLeftMargin(); cout << "  scale     " << radius << " " << radius << " " << radius << endl;
  double rotationAngle = (int) floor(360.0 * RNG::uniform());
  Vector3D rotationAxis(-1.0 + (2.0 * RNG::uniform()),
                        -1.0 + (2.0 * RNG::uniform()),
                        -1.0 + (2.0 * RNG::uniform()));
  rotationAxis.normalize();
  makeLeftMargin(); cout << "  rotate    " << rotationAngle << " "
                                           << rotationAxis.fX << " "
                                           << rotationAxis.fY << " "
                                           << rotationAxis.fZ << endl;
  makeLeftMargin(); cout << "  sphere" << endl;
  makeLeftMargin(); cout << "pop" << endl;
}

void recursivelyGenerateSpheres(double mx,
                                double my,
                                double mz, 
                                double radius,
                                double inhibitionFactor,
                                double offsetFactor,
                                int generationsLeft)
{
  generateOneSphere(mx,my,mz,radius);

  --generationsLeft;

  if (generationsLeft != 0) {

    double newRadius = (radius * inhibitionFactor);
    double offset = radius + (newRadius * offsetFactor);

    recursivelyGenerateSpheres(mx + offset,my         ,mz         ,newRadius,inhibitionFactor,offsetFactor,generationsLeft);
    recursivelyGenerateSpheres(mx         ,my         ,mz - offset,newRadius,inhibitionFactor,offsetFactor,generationsLeft);
    recursivelyGenerateSpheres(mx - offset,my         ,mz         ,newRadius,inhibitionFactor,offsetFactor,generationsLeft);
    recursivelyGenerateSpheres(mx         ,my         ,mz + offset,newRadius,inhibitionFactor,offsetFactor,generationsLeft);
    recursivelyGenerateSpheres(mx         ,my - offset,mz         ,newRadius,inhibitionFactor,offsetFactor,generationsLeft);
    recursivelyGenerateSpheres(mx         ,my + offset,mz         ,newRadius,inhibitionFactor,offsetFactor,generationsLeft);
  }
}

void main(int argc, char* argv[])
{
  if (argc != (4 + 1)) {
    cout << endl;
    cout << "Usage : generate-fractal-spheres <left-margin> <inhibition-factor> <offset-factor>" << endl;
    cout << "                                 <#generations>" << endl;
    cout << endl;
    exit(0);
  }

  kLeftMargin = (int) getValue(argv[1]);

  RNG::setSeedWithTime();

  recursivelyGenerateSpheres(0.0,0.0,0.0,1.0,
                             getValue(argv[2]),
                             (int) getValue(argv[3]),
                             (int) getValue(argv[4]));
}
