// ------------------------------------------------------
// Filename      : generate-quadric-reflections-frame.cpp
// Author        : Sven Maerivoet
// Last modified : 30/08/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------------------------------

#include <iostream>
#include <math.h>
#include <string>
#include "opengl/math.h"

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  stringstream strstream
#endif

using namespace std;

double getValue(string str)
{
  stringstream tmp;
  double number;

  tmp << str;
  if (!(tmp >> number)) {
    cerr << "getValue() -> error getting value [" << str << "]." << endl;
    exit(1);
  }
  else {
    char t;
    if (tmp >> t) {
      cerr << "getValue() -> bum chars in number [" << str << "]." << endl;
      exit(1);
    } 
  }

  return number;
}

void generateQuadricReflectionsFrame(int part,
                                     int framesPerPart,
                                     int currentFrame)
{
  cout << "background     0.2 0.1  0.3" << endl;
  cout << "globalAmbient  0.2 0.2  0.2" << endl;

  double minX, minY, minZ,
         maxX, maxY, maxZ,
         x, y, z,
         turbulence,
         reflectivity,
         glossStrength;
  double fraction = ((double) (currentFrame - 1)) / ((double) (framesPerPart - 1));

  switch (part) {
    case 1 :
    case 2 :
    case 3 :
    case 4 :
      cout << "light          5.0 5.0 -2.0 1.0 1.0 1.0" << endl;
      break;

    case 5 : 
      minX = 5.0;
      maxX = -5.0;
      x = Math::lerp(minX,maxX,fraction);
      cout << "light          " << x << " 5.0 -2.0 1.0 1.0 1.0" << endl;
      break;

    case 6 : 
    case 7 : 
      cout << "light         -5.0 5.0 -2.0 1.0 1.0 1.0" << endl;
      break;
  }

  cout << "light          2.0 1.5  2.0 0.4 0.4 0.4" << endl;
  cout << endl;
  cout << "include data/sdl/include/colors.sdl" << endl;
  cout << endl;

  switch (part) {
    case 1 :
      minX = 4.9;
      minY = 3.7;
      minZ = 4.8;
      maxX = 2.4;
      maxY = 1.8;
      maxZ = 2.3;
      x = Math::lerp(minX,maxX,fraction);
      y = Math::lerp(minY,maxY,fraction);
      z = Math::lerp(minZ,maxZ,fraction);
      cout << "eyepoint  " << x << " " << y << " " << z << endl;
      break;

    case 2 :
      cout << "eyepoint  2.4 1.8 2.3" << endl;
      break;

    case 3 :
      minX = 2.4;
      minY = 1.8;
      minZ = 2.3;
      maxX = 1.3;
      maxY = 1.0;
      maxZ = 1.3;
      x = Math::lerp(minX,maxX,fraction);
      y = Math::lerp(minY,maxY,fraction);
      z = Math::lerp(minZ,maxZ,fraction);
      cout << "eyepoint  " << x << " " << y << " " << z << endl;
      break;

    case 4 :
    case 5 :
    case 6 :
    case 7 :
      cout << "eyepoint  1.3 1.0 1.3" << endl;
      break;
  }
  cout << "viewpoint 0.0 0.0 0.0" << endl;

  cout << endl;
  cout << "showExtents 1" << endl;
  cout << endl;
  cout << "atmosphericAttenuation 1.0 1.0" << endl;
  cout << "                       7.0 0.0" << endl;
  cout << "                       0.2 0.1 0.3" << endl;
  cout << endl;
  cout << "push" << endl;

  switch (part) {
    case 1 :
    case 2 :
    case 3 :
      cout << "  use                ColorWhite" << endl;
      cout << "  reflectivity       1.0" << endl;
      cout << "  solidTexture       8   2    0.0  4.0 5.0 6.0" << endl;
      break;

    case 4 :
      cout << "  use                ColorWhite" << endl;
      cout << "  reflectivity       1.0" << endl;
      turbulence = Math::lerp(0.0,5.0,fraction);
      cout << "  solidTexture       8   2    " << turbulence << " 4.0 5.0 6.0" << endl;
      break;

    case 5 :
      cout << "  use                ColorRed" << endl;
      cout << "  reflectivity       1.0" << endl;
      cout << "  solidTexture       1   0.4 0.8 0.4" << endl;
      break;

    case 6 :
      cout << "  use                ColorRed" << endl;
      reflectivity = Math::lerp(1.0,0.0,fraction);
      cout << "  reflectivity       " << reflectivity << endl;
      cout << "  solidTexture       1   0.4 0.8 0.4" << endl;
      break;

    case 7 :
      cout << "  use                ColorRed" << endl;
      cout << "  solidTexture       1   0.4 0.8 0.4" << endl;
      break;
  }

  cout << "  translate          0.0 0.0 -0.5" << endl;
  cout << "  scale              0.5 0.5  0.5" << endl;
  cout << "  ellipticParaboloid 2.0 2.0  5.0" << endl;
  cout << "pop" << endl;
  cout << endl;
  cout << "defaultMaterials" << endl;
  cout << endl;
  cout << "push" << endl;
  cout << "  use           ColorYellow" << endl;

  switch (part) {
    case 1 :
      cout << "  translate     2.5 1.0 0.5" << endl;
      break;

    case 2 :
      minX = 2.5;
      minY = 1.0;
      minZ = 0.5;
      maxX = 1.0;
      maxY = 0.5;
      maxZ = 0.5;
      x = Math::lerp(minX,maxX,fraction);
      y = Math::lerp(minY,maxY,fraction);
      z = Math::lerp(minZ,maxZ,fraction);
      cout << "  translate     " << x << " " << y << " " << z << endl;
      break;

    case 3 :
    case 4 :
    case 5 :
      cout << "  translate     1.0 0.5 0.5" << endl;
      break;

    case 6 :
      reflectivity = Math::lerp(0.0,1.0,fraction);
      cout << "  reflectivity  " << reflectivity << endl;
      cout << "  translate     1.0 0.5 0.5" << endl;
      break;

    case 7 :
      glossStrength = Math::lerp(0.0,0.5,fraction);
      cout << "  reflectivity  1.0" << endl;
      cout << "  glossStrength " << glossStrength << endl;
      cout << "  translate     1.0 0.5 0.5" << endl;
      break;
  }
  cout << "  scale         0.3 0.3 0.3" << endl;
  cout << "  sphere" << endl;
  cout << "pop" << endl;
}

void main(int argc, char* argv[])
{
  if (argc != (1 + 3)) {
    cout << endl;
    cout << "Usage : generate-quadric-reflections-frame <part> <frames-per-part> <current-frame>" << endl;
    cout << endl;
    cout << "Part 1 zooms in on the quadric, part 2 shows the yellow ball," << endl;
    cout << "part 3 zooms further in, part 4 increases the quadric's turbulence," << endl;
    cout << "part 5 changes the position of a light, part 6 changes the reflectivities and" << endl;
    cout << "part 7 changes the gloss-strength." << endl;
    cout << endl;
    exit(0);
  }

  generateQuadricReflectionsFrame((int) getValue(argv[1]),(int) getValue(argv[2]),(int) getValue(argv[3]));
}
