// ------------------------------
// Filename      : affine.h
// Author        : Sven Maerivoet
// Last modified : 30/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_AFFINE_H // include-guard
#define SM_AFFINE_H

#ifdef _WIN32
  #include <windows.h>
#endif

#include "primitives.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <vector>

using namespace std;

class AffineTransformation {
  public:
    // constructors
    AffineTransformation(void);
    AffineTransformation(const AffineTransformation& affineTransformation);

    // destructor
    ~AffineTransformation(void);

    // methods
    AffineTransformation& operator=(const AffineTransformation& affineTransformation);
    void set(const AffineTransformation& affineTransformation);
    void setIdentityMatrix(void);

    void preMult(const AffineTransformation& affineTransformation);
    void postMult(const AffineTransformation& affineTransformation);

    AffineTransformation operator*(double scalar) const;
    AffineTransformation& operator*=(double scalar);

    static AffineTransformation convertRotationToTransformation(double angleInDegrees, const Vector3D& u);
    static AffineTransformation convertScalingToTransformation(double sx, double sy, double sz);
    static AffineTransformation convertTranslationToTransformation(const Vector3D& direction);

    // data-members
    double m[16]; // 4x4 matrix
};

class AffineTransformationNode {
  public:
    // constructor
    AffineTransformationNode(void);

    // destructor
    ~AffineTransformationNode(void);

    // data-members
    AffineTransformation*     fAffineTransformation;
    AffineTransformation*     fInverseAffineTransformation;
    AffineTransformationNode* fNext;
};

class AffineTransformationStack {
  public:
    // constructor
    AffineTransformationStack(void);

    // destructor
    ~AffineTransformationStack(void);

    // methods
    void duplicate(void);
    void setIdentity(void);
    void pop(void);
    void clear(void);
    void rotate(double angleInDegrees, const Vector3D& u);
    void scale(double sx, double sy, double sz);
    void translate(const Vector3D& direction);
    void transform(const Vector3D& row1,
                   const Vector3D& row2,
                   const Vector3D& row3);

    // data-members
    AffineTransformationNode* fTopOfStack;
};

#endif
