// ------------------------------
// Filename      : chrono.cpp
// Author        : Sven Maerivoet
// Last modified : 20/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "chrono.h"
#include <string>
#include <time.h>

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  ostringstream ostrstream
#endif

using namespace std;

// ------------
// class Chrono
// ------------

// -----------
// constructor
// -----------
Chrono::Chrono(void) : fRunning(false),
                       fElapsedTime(0),
                       fStartTime(0)
{
}

// -----------
// destructor
// -----------
Chrono::~Chrono(void)
{
}

// -------
// methods
// -------
void Chrono::start(void)
{
  time(&fStartTime);
  fRunning = true;
}

void Chrono::stop(void)
{
  if (fRunning) {
    time(&fElapsedTime);
    fElapsedTime = fElapsedTime - fStartTime;
    fRunning = false;
  }
}

long Chrono::getElapsedTime(void)
{
  if (fRunning) {
    time(&fElapsedTime);
    fElapsedTime = fElapsedTime - fStartTime;
  }

  return fElapsedTime;
}

string Chrono::convertToHMS(long elapsedTime)
{
  int hours, minutes, seconds;

  hours = elapsedTime / (60 * 60);
  elapsedTime = elapsedTime % (60 * 60);

  minutes = elapsedTime / 60;
  elapsedTime = elapsedTime % 60;

  seconds = elapsedTime;

  string HMS = "";
  if (hours < 10) {
    HMS += "0";
  }
  HMS += convertIntToString(hours) + ":";

  if (minutes < 10) {
    HMS += "0";
  }
  HMS += convertIntToString(minutes) + ":";

  if (seconds < 10) {
    HMS += "0";
  }
  HMS += convertIntToString(seconds);

  return HMS;
}

// ---------------
// utility methods
// ---------------
string Chrono::convertIntToString(int x)
{
  ostringstream ostr;
  ostr << x;

  // terminate stream
  #ifndef _WIN32
    ostr << ends; // Microsoft Visual C++ 6.0 puts a strange symbol here
  #endif

  return ostr.str();
}
