// ------------------------------
// Filename      : extents.h
// Author        : Sven Maerivoet
// Last modified : 25/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_EXTENTS_H // include-guard
#define SM_EXTENTS_H

#ifdef _WIN32
  #include <windows.h>
#endif

#include "affine.h"
#include "constants.h"
#include "primitives.h"
#include "ray.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

using namespace std;

class PointCluster {
  public:
    // constructors
    PointCluster(void);
    PointCluster(int nrOfPoints);

    // destructor
    ~PointCluster(void);

    // methods
    Point3D calcCentroid(void) const;
    void applyAffineTransformation(const AffineTransformation& affine);
    void applyTranslation(const AffineTransformation& affine);

    // data-members
    int      fNrOfPoints;
    Point3D* fPoints;
};

class GenericExtent {
  public:
    // constructor
    GenericExtent(void);

    // destructor
    ~GenericExtent(void);

    // methods
    virtual void create(const PointCluster& pointCluster, double excess);
    virtual bool rayIntersectsExtent(const Ray& ray) const;
    virtual void draw(void) const;
};

class BoxExtent : public GenericExtent {
  public:
    // constructor
    BoxExtent(void);

    // destructor
    ~BoxExtent(void);

    // methods
    void create(const PointCluster& pointCluster, double excess);
    bool rayIntersectsExtent(const Ray& ray) const;
    void draw(void) const;

    // data-members
    double fLeft;
    double fRight;
    double fBottom;
    double fTop;
    double fBack;
    double fFront;
};

class SphereExtent : public GenericExtent {
  public:
    // constructor
    SphereExtent(void);

    // destructor
    ~SphereExtent(void);

    // methods
    void create(const PointCluster& pointCluster, double excess);
    bool rayIntersectsExtent(const Ray& ray) const;
    void draw(void) const;

    // data-members
    Point3D fOrigin;
    double  fRadius;
};

#endif
