// ------------------------------
// Filename      : math.h
// Author        : Sven Maerivoet
// Last modified : 18/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_MATH_H // include-guard
#define SM_MATH_H

class Math {
  public:
    // useful constants
    static const double kPi;
    static const double kEpsilon;
    static const double kInfinity;

    // methods
    static double abs(double x);
    static double frac(double x);
    static double sqr(double x);
    static double cube(double x);
    static double quart(double x);
    static double cubicRoot(double x);
    static double lerp(double a, double b, double fraction);
    static double arctan(double y, double x);
    static double radToDeg(double radians);
    static double degToRad(double degrees);
    static void convertCartesianToSpherical(double x, double y, double z,
                                            double& radius, double& phi, double& theta);
    static int solveQuadraticEquation(double coefficients[3], double solutions[2]);
    static int solveCubicEquation(double coefficients[4], double solutions[3]);
    static int solveQuarticEquation(double coefficients[5], double solutions[4]);

  private:
    static const double kSolverEpsilon;
};

#endif
