// ------------------------------
// Filename      : rgbapixmap.h
// Author        : Sven Maerivoet
// Last modified : 10/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_RGBAPIXMAP_H // include-guard
#define SM_RGBAPIXMAP_H

#include <string>
#include <fstream>

using namespace std;

// some typedefs for convenience
typedef unsigned char  uchar;
typedef unsigned short ushort;
typedef unsigned long  ulong;

class mRGBA {
  public:
    // constructors
    mRGBA(void);
    mRGBA(const mRGBA& mrgba);
    mRGBA(uchar red, uchar green, uchar blue);
    mRGBA(uchar red, uchar green, uchar blue, uchar alpha);

    // destructor
    ~mRGBA(void);

    // methods
    void set(uchar red, uchar green, uchar blue);
    void set(uchar red, uchar green, uchar blue, uchar alpha);
    void convert(double& red, double& green, double& blue) const;
    double normalizeAlpha(void) const;

    // data-members
    uchar fRed, fGreen, fBlue, fAlpha;
};

class RGBApixmap {
  public:
    // constructors
    RGBApixmap(void);
    RGBApixmap(int row, int cols);

    // destructor
    ~RGBApixmap(void);

    // methods
    void clear(void);
    bool loadFromBMPFile(string filename);
    bool writeToBMPFile(string filename);

    void setTextureID(int textureID);
    int getTextureID(void) const;

    int getNrOfRows(void) const;
    int getNrOfColumns(void) const;

    void setRGBATriplet(int u, int v, const mRGBA& mrgba);
    mRGBA getRGBATriplet(int u, int v) const;

    void getColor(int u, int v, double& red, double& green, double& blue) const;
    mRGBA* getPixmap(void) const;

    bool copyFromScreen(int left, int bottom,
                        int width, int height);
    void drawToScreen(int left, int bottom) const;
    void copyFromScreenToScreen(int left, int bottom,
                                int width, int height,
                                int newLeft, int newBottom) const;

    void setChromaKey(const mRGBA& chromaKey);
    void blendOnScreen(int left, int bottom) const;

  private:
    // utility methods
    ushort readShort(void);
    ulong readLong(void);
    void writeShort(ushort us);
    void writeLong(ulong ul);
    
    // data-members
    int     fNRows, fNCols;
    mRGBA*  fPixmap; // array of pixels
    fstream fFile;
    int     fTextureID;
};

#endif
