// ------------------------------
// Filename      : scene.cpp
// Author        : Sven Maerivoet
// Last modified : 16/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "math.h"
#include "rgbapixmap.h"
#include "rng.h"
#include "scene.h"
#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>

using namespace std;

// -----------
// class Light
// -----------

// -----------
// constructor
// -----------
Light::Light(void) : fPosition(),
                     fColor(),
                     fIsSpotLight(false),
                     fSpotLightDirection(),
                     fCutOffAngle(0.0),
                     fExponent(0.0)
{
}

// ---------
// destructor
// ----------
Light::~Light(void)
{
}

// -------------
// class DefUnit
// -------------

// -----------
// constructor
// -----------
DefUnit::DefUnit(string name, string stuff) : fName(name),
                                              fStuff(stuff)
{
}

// ----------
// destructor
// ----------
DefUnit::~DefUnit(void)
{
}

// ------------------
// class DefUnitStack
// ------------------

// -----------
// constructor
// -----------
DefUnitStack::DefUnitStack(void) : fStack(0)
{
}

// ----------
// destructor
// ----------
DefUnitStack::~DefUnitStack(void)
{
  release();
}

// -------
// methods
// -------
void DefUnitStack::push(string name, string stuff)
{
  D4S* tempD4S = new D4S();

  tempD4S->fCurrent = new DefUnit(name,stuff);
  tempD4S->fNext = fStack;
  fStack = tempD4S;
}

bool DefUnitStack::search(string searchName) const
{
  string toSearch = searchName;

  // convert searchName to it's uppercase equivalent
  for (int i = 0; i < toSearch.length(); ++i) {
    if ((((int) toSearch[i]) >= 97) &&
        (((int) toSearch[i]) <= 122)) {
      toSearch[i] = toupper(toSearch[i]);
    }
  }

  D4S* temp = fStack;

  while (temp != 0) {

    string name = temp->fCurrent->fName;

    // convert fName to it's uppercase equivalent
    for (int i = 0; i < name.length(); ++i) {
      if ((((int) name[i]) >= 97) &&
          (((int) name[i]) <= 122)) {
        name[i] = toupper(name[i]);
      }
    }

    if (name == toSearch) {
      return true;
    }
    temp = temp->fNext;
  }
  return false;
}

string DefUnitStack::contents(string searchName) const
{
  string toSearch = searchName;

  // convert searchName to it's uppercase equivalent
  for (int i = 0; i < toSearch.length(); ++i) {
    if ((((int) toSearch[i]) >= 97) &&
        (((int) toSearch[i]) <= 122)) {
      toSearch[i] = toupper(toSearch[i]);
    }
  }

  D4S* temp = fStack;

  while (temp != 0) {

    string name = temp->fCurrent->fName;

    // convert fName to it's uppercase equivalent
    for (int i = 0; i < name.length(); ++i) {
      if ((((int) name[i]) >= 97) &&
          (((int) name[i]) <= 122)) {
        name[i] = toupper(name[i]);
      }
    }

    if (name == toSearch) {
      return(temp->fCurrent->fStuff);
    }
    temp = temp->fNext;
  }
  return 0;
}

void DefUnitStack::release(void)
{
  while (fStack != 0) {
    D4S* tmp = fStack;

    fStack = fStack->fNext;
    delete tmp->fCurrent;
    delete tmp;
  }
  fStack = 0;
}

void DefUnitStack::add(DefUnitStack* defUnitStack)
{
  D4S* tmp = defUnitStack->fStack;

  while (tmp != 0) {

    string name = tmp->fCurrent->fName;
    string stuff = tmp->fCurrent->fStuff;
    push(name,stuff);

    tmp = tmp->fNext;
  }
}

// ----------
// class Line
// ----------

// -----------
// constructor
// -----------
Line::Line(void)
{
}

// ----------
// destructor
// ----------
Line::~Line(void)
{
}

// -----------
// class Scene
// -----------

// ------------
// constructors
// ------------
Scene::Scene(void)
{
  initialize();
}

Scene::Scene(string filename)
{
  readFromFile(filename);
}

// ----------
// destructor
// ----------
Scene::~Scene(void)
{
  delete fLights;
  fLights = 0;
  delete fGeometricObjects;
  fGeometricObjects = 0;

  if (fDefStack != 0) {
    delete fDefStack;
    fDefStack = 0;
  }
}

// -------
// methods
// -------
void Scene::clear(void)
{
  fLights->clear();
  fGeometricObjects->clear();
}

void Scene::makeLightsOpenGL(void) const
{
  // handle the global ambient light
  GLfloat globalAmbientColor[4];
  fGlobalAmbientColor.build4tuple(globalAmbientColor);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,globalAmbientColor);

  int lightNr = GL_LIGHT0;
  GLfloat color[4], position[4];

  Lights::const_iterator light;

  for (light = fLights->begin(); light != fLights->end(); ++light) {
    (*light)->fColor.build4tuple(color);
    (*light)->fPosition.build4tuple(position);

    GLfloat zero[] = {(GLfloat) 0.0,
                      (GLfloat) 0.0,
                      (GLfloat) 0.0,
                      (GLfloat) 1.0};

    glLightfv((GLenum) lightNr,GL_AMBIENT,zero);
    glLightfv((GLenum) lightNr,GL_DIFFUSE,color);
    glLightfv((GLenum) lightNr,GL_SPECULAR,color);
    glLightfv((GLenum) lightNr,GL_POSITION,position);

    // setup spot-light
    if ((*light)->fIsSpotLight) {
      GLfloat spotLightDirection[4];
      (*light)->fSpotLightDirection.build4tuple(spotLightDirection);
      glLightfv((GLenum) lightNr,GL_SPOT_DIRECTION,spotLightDirection);
      glLightf((GLenum) lightNr,GL_SPOT_CUTOFF,(GLfloat) ((*light)->fCutOffAngle));
      glLightf((GLenum) lightNr,GL_SPOT_EXPONENT,(GLfloat) ((*light)->fExponent));
    }

    // compute light-source attenuation
    glLightf((GLenum) lightNr,GL_CONSTANT_ATTENUATION,(GLfloat) fLightAttenuationConstantFactor);
    glLightf((GLenum) lightNr,GL_LINEAR_ATTENUATION,(GLfloat) fLightAttenuationLinearFactor);
    glLightf((GLenum) lightNr,GL_QUADRATIC_ATTENUATION,(GLfloat) fLightAttenuationQuadraticFactor);

    glEnable((GLenum) lightNr);
    ++lightNr;
  }

  // compute the fog
  if (fFogStart != fFogEnd) {
    GLfloat fogColor4[] = {(GLfloat) fFogColor.fRed,
                           (GLfloat) fFogColor.fGreen,
                           (GLfloat) fFogColor.fBlue,
                           (GLfloat) 1.0};
    glFogfv(GL_FOG_COLOR,fogColor4);
    glHint(GL_FOG_HINT,GL_DONT_CARE);

    glFogi(GL_FOG_MODE,GL_LINEAR);
    glFogf(GL_FOG_START,(GLfloat) fFogStart);
    glFogf(GL_FOG_END,(GLfloat) fFogEnd);

    glEnable(GL_FOG);
  }
}

void Scene::drawOpenGL(bool depthBufferEnabled) const
{
  glClearColor((GLclampf) fBackgroundColor.fRed,
               (GLclampf) fBackgroundColor.fGreen,
               (GLclampf) fBackgroundColor.fBlue,
               (GLclampf) 1.0);

  if (depthBufferEnabled) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  }
  else {
    glClear(GL_COLOR_BUFFER_BIT);
  }

  GeometricObjects::const_iterator p;
  for (p = fGeometricObjects->begin(); p != fGeometricObjects->end(); ++p) {
    (*p)->tellMaterialsGL();
    (*p)->drawOpenGL();
    if (fShowExtents) {

      glColor4d((GLdouble) 1.0,
                (GLdouble) 1.0,
                (GLdouble) 1.0,
                (GLdouble) 1.0);

      GLfloat black[]     = {(GLfloat) 0.0,
                             (GLfloat) 0.0,
                             (GLfloat) 0.0,
                             (GLfloat) 1.0};
      GLfloat white[]     = {(GLfloat) 1.0,
                             (GLfloat) 1.0,
                             (GLfloat) 1.0,
                             (GLfloat) 1.0};
      GLfloat shininess[] = {(GLfloat) 0.0};

      glMaterialfv(GL_FRONT,GL_AMBIENT,black);
      glMaterialfv(GL_FRONT,GL_DIFFUSE,black);
      glMaterialfv(GL_FRONT,GL_SPECULAR,black);
      glMaterialfv(GL_FRONT,GL_SHININESS,shininess);
      glMaterialfv(GL_FRONT,GL_EMISSION,white);

      (*p)->drawExtent();
    }
  }
}

void Scene::readFromFile(string filename)
{
  initialize();

  fSceneLoaded = false;

  ifstream* fFileInputStream = new ifstream(filename.c_str());
  if (!(*fFileInputStream)) {
    cerr << "Scene::readFromFile() -> can't find or open file [" << filename << "]." << endl; 
    return;
  }

  cerr << "Loading SDL-file [" << filename << "]..." << endl;

  clear();
  fCurrentLine = 1;
  fNextLine = fCurrentLine;
  fDefStack = new DefUnitStack();

  fAffineStack.fTopOfStack = new AffineTransformationNode();
  fAffineStack.fTopOfStack->fNext = 0;

  fMotionBlurAffineStack.fTopOfStack = new AffineTransformationNode();
  fMotionBlurAffineStack.fTopOfStack->fNext = 0;

  // buffer file completely
  char ch;
  fInputStream = new stringstream();
  while (fFileInputStream->get(ch)) {
    *fInputStream << ch;
  }

  bool ok = true;
  int nrOfGeometricObjects = 0;
  while (ok) {
    GeometricObject* shape = getGeometricObject();

    if (shape != 0) {
      fGeometricObjects->push_back(shape);
      ++nrOfGeometricObjects;
    }
    else {
      ok = false;
    }
  } 
  cerr << "  -> " << nrOfGeometricObjects << " geometric objects loaded for [" << filename << "]." << endl;

  fFileInputStream->close();
  fAffineStack.clear();
  fMotionBlurAffineStack.clear();
  fSceneLoaded = true;
}

bool Scene::sceneLoaded(void) const
{
  return fSceneLoaded;
}

void Scene::addLight(Light* light)
{
  fLights->push_back(light);
}

void Scene::addGeometricObject(GeometricObject* geometricObject)
{
  fGeometricObjects->push_back(geometricObject);
}

Color Scene::shootRay(const Ray& ray)
{
  // update statistics
  ++fNrOfRaysShot;

  // find the intersection of the ray with the nearest object
  Intersection intersection;
  Intersection bestIntersection;

  bestIntersection.fNrOfHits = 0;

  GeometricObjects::const_iterator p;
  for (p = fGeometricObjects->begin(); p != fGeometricObjects->end(); ++p) {

    if (!(*p)->computeIntersectionsWithRay(ray,intersection)) {
      // the ray didn't hit this object
      continue;
    }

    if ((bestIntersection.fNrOfHits == 0) ||
        (intersection.fHits[0].fHitTime < bestIntersection.fHits[0].fHitTime)) {
      bestIntersection.set(intersection);
    }
  }

  if (bestIntersection.fNrOfHits == 0) {

    if (fFollowRays && ray.fFollowRay) {
      Line line;
      line.fp1 = ray.fStart;
      line.fp2 = Point3D(ray.fStart.fX + ray.fDirection.fX,
                         ray.fStart.fY + ray.fDirection.fY,
                         ray.fStart.fZ + ray.fDirection.fZ);
      // use a different color to show that this ray didn't hit any object
      line.fColor = Color(0.5,0.5,0.5);

      fLines.push_back(line);
    }
   
    // the ray didn't hit any object
    if (ray.fIsReflective) {
      // prevent irrealistic reflection of the background
      // note that a ray is either reflective, refractive or neither
      return Color(0.0,0.0,0.0);
    }
    else {
      // note that the background is visible through transparent objects
      return fBackgroundColor;
    }
  }

  HitInfo hitInfo;
  hitInfo.set(bestIntersection.fHits[0]);

  // calculate the point where the ray hit the object
  Point3D hitPoint(ray.fStart.fX + (ray.fDirection.fX * hitInfo.fHitTime),
                   ray.fStart.fY + (ray.fDirection.fY * hitInfo.fHitTime),
                   ray.fStart.fZ + (ray.fDirection.fZ * hitInfo.fHitTime));

  Shape* objectHit = (Shape*) hitInfo.fObjectHit;

  // setup the normal at the point where the ray hit the object
  Vector3D normal;
  Transformations::transformNormal(normal,
                                   objectHit->fInverseAffineTransformation,
                                   hitInfo.fHitNormal);
  normal.normalize();

  // randomly perturb the normal according to the object's surface-roughness
  if (objectHit->fMaterial.fSurfaceRoughness != 0.0) {
    normal.fX += -(objectHit->fMaterial.fSurfaceRoughness / 2.0) +
                  (objectHit->fMaterial.fSurfaceRoughness * RNG::uniform());
    normal.fY += -(objectHit->fMaterial.fSurfaceRoughness / 2.0) +
                  (objectHit->fMaterial.fSurfaceRoughness * RNG::uniform());
    normal.fZ += -(objectHit->fMaterial.fSurfaceRoughness / 2.0) +
                  (objectHit->fMaterial.fSurfaceRoughness * RNG::uniform());
    normal.normalize();
  }

  if (fFindNormals) {

    if (RNG::uniform() < fFindNormalsProbability) {
      Line line;
      line.fp1 = hitPoint;
      line.fp2 = Point3D(hitPoint.fX + normal.fX,
                         hitPoint.fY + normal.fY,
                         hitPoint.fZ + normal.fZ);

      // let the line's color have a fine blue color
      line.fColor = Color(0.0,0.0,1.0);

      fLines.push_back(line);
    }
  }

  if (fFollowRays && ray.fFollowRay) {
    Line line;
    line.fp1 = ray.fStart;
    line.fp2 = hitPoint;
    line.fColor = ray.fColor;

    fLines.push_back(line);
  }

  // setup a shadowfeeler (use a small offset to prevent self-shadowing)
  Vector3D normalizedHitDirection(ray.fDirection);
  normalizedHitDirection.normalize();
  Ray shadowFeeler(Point3D(hitPoint.fX - (fShadowFeelerEpsilon * normalizedHitDirection.fX),
                           hitPoint.fY - (fShadowFeelerEpsilon * normalizedHitDirection.fY),
                           hitPoint.fZ - (fShadowFeelerEpsilon * normalizedHitDirection.fZ)));
  shadowFeeler.fNrOfConeSamples = ray.fNrOfConeSamples;

  // compute the final color at the hitPoint (start with black)
  Color color(0.0,0.0,0.0);

  // handle emissive color
  color = Color(objectHit->fMaterial.fEmissive);

  bool solidTextureEnabled = true;
  Color solidTextureModulation(1.0,1.0,1.0);

  if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kCheckerboard) {
    solidTextureModulation = fSolidTexturer.sampleCheckerboard(hitInfo.fHitPoint,
                                                               Vector3D(objectHit->fMaterial.fSolidTextureParameters[0],
                                                                        objectHit->fMaterial.fSolidTextureParameters[1],
                                                                        objectHit->fMaterial.fSolidTextureParameters[2]));
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kStackedColorBars) {
    solidTextureModulation = fSolidTexturer.sampleStackedColorBars(hitInfo.fHitPoint,
                                                                   Vector3D(objectHit->fMaterial.fSolidTextureParameters[0],
                                                                            objectHit->fMaterial.fSolidTextureParameters[1],
                                                                            objectHit->fMaterial.fSolidTextureParameters[2]));
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kRGBCubes) {
    solidTextureModulation = fSolidTexturer.sampleRGBCubes(hitInfo.fHitPoint,
                                                           Vector3D(objectHit->fMaterial.fSolidTextureParameters[0],
                                                                    objectHit->fMaterial.fSolidTextureParameters[1],
                                                                    objectHit->fMaterial.fSolidTextureParameters[2]));
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kOneCoordinateBasedContourLines) {
    solidTextureModulation = fSolidTexturer.sampleOneCoordinateBasedContourLines(hitInfo.fHitPoint,
                                                                                 objectHit->fMaterial.fSolidTextureParameters[0],
                                                                                 objectHit->fMaterial.fSolidTextureParameters[1]);
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kTwoCoordinatesBasedContourLines) {
    solidTextureModulation = fSolidTexturer.sampleTwoCoordinateBasedContourLines(hitInfo.fHitPoint,
                                                                                 objectHit->fMaterial.fSolidTextureParameters[0],
                                                                                 objectHit->fMaterial.fSolidTextureParameters[1],
                                                                                 Vector2D(objectHit->fMaterial.fSolidTextureParameters[2],
                                                                                          objectHit->fMaterial.fSolidTextureParameters[3]));
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kThreeCoordinatesBasedContourLines) {
    solidTextureModulation = fSolidTexturer.sampleThreeCoordinateBasedContourLines(hitInfo.fHitPoint,
                                                                                   objectHit->fMaterial.fSolidTextureParameters[0],
                                                                                   Vector3D(objectHit->fMaterial.fSolidTextureParameters[1],
                                                                                            objectHit->fMaterial.fSolidTextureParameters[2],
                                                                                            objectHit->fMaterial.fSolidTextureParameters[3]));
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kWoodGrain) {
    solidTextureModulation = fSolidTexturer.sampleWoodGrain(hitInfo.fHitPoint,
                                                            objectHit->fMaterial.fSolidTextureParameters[0],
                                                            objectHit->fMaterial.fSolidTextureParameters[1],
                                                            objectHit->fMaterial.fSolidTextureParameters[2],
                                                            objectHit->fMaterial.fSolidTextureParameters[3],
                                                            Color(objectHit->fMaterial.fSolidTextureParameters[4],
                                                                  objectHit->fMaterial.fSolidTextureParameters[5],
                                                                  objectHit->fMaterial.fSolidTextureParameters[6]),
                                                            Color(objectHit->fMaterial.fSolidTextureParameters[7],
                                                                  objectHit->fMaterial.fSolidTextureParameters[8],
                                                                  objectHit->fMaterial.fSolidTextureParameters[9]),
                                                            (objectHit->fMaterial.fSolidTextureParameters[10] != 0.0));
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kMarble) {
    solidTextureModulation = fSolidTexturer.sampleMarble(hitInfo.fHitPoint,
                                                         objectHit->fMaterial.fSolidTextureParameters[0],
                                                         objectHit->fMaterial.fSolidTextureParameters[1],
                                                         objectHit->fMaterial.fSolidTextureParameters[2],
                                                         objectHit->fMaterial.fSolidTextureParameters[3],
                                                         objectHit->fMaterial.fSolidTextureParameters[4]);
  }
  else if (objectHit->fMaterial.fSolidTextureType == SolidTexturer::kCowSpots) {
    solidTextureModulation = fSolidTexturer.sampleCowSpots(hitInfo.fHitPoint,
                                                           objectHit->fMaterial.fSolidTextureParameters[0]);
  }
  else {
    solidTextureEnabled = false;
  }

  bool flatTextureEnabled = true;
  Color flatTextureModulation(1.0,1.0,1.0);

  if ((objectHit->fMaterial.fHasImageTexture) ||
      (objectHit->fMaterial.fProceduralTextureType != ProceduralTexturer::kNone)) {

    double s, t;
    objectHit->computeTextureCoordinates(hitInfo,s,t);

    if (objectHit->fMaterial.fHasImageTexture) {

      double textureS = s * (objectHit->fMaterial.fImageTextureBitmap->getNrOfColumns() - 2);
      double textureT = t * (objectHit->fMaterial.fImageTextureBitmap->getNrOfRows() - 2);
      
      int indexS = (int) floor(textureS);
      int indexT = (int) floor(textureT);

      double fractionS = Math::frac(textureS);
      double fractionT = Math::frac(textureT);

      // interpolate the color between the four corners that surround the texture-point
      Color cUpperLeft, cUpperRight, cLowerLeft, cLowerRight;
      objectHit->fMaterial.fImageTextureBitmap->getColor(textureS,textureT,
                                                         cUpperLeft.fRed,cUpperLeft.fGreen,cUpperLeft.fBlue);
      objectHit->fMaterial.fImageTextureBitmap->getColor(textureS + 1,textureT,
                                                         cUpperRight.fRed,cUpperRight.fGreen,cUpperRight.fBlue);
      objectHit->fMaterial.fImageTextureBitmap->getColor(textureS,textureT + 1,
                                                         cLowerLeft.fRed,cLowerLeft.fGreen,cLowerLeft.fBlue);
      objectHit->fMaterial.fImageTextureBitmap->getColor(textureS + 1,textureT + 1,
                                                         cLowerRight.fRed,cLowerRight.fGreen,cLowerRight.fBlue);

      Color cUpper(Math::lerp(cUpperLeft.fRed,cUpperRight.fRed,fractionS),
                   Math::lerp(cUpperLeft.fGreen,cUpperRight.fGreen,fractionS),
                   Math::lerp(cUpperLeft.fBlue,cUpperRight.fBlue,fractionS));

      Color cLower(Math::lerp(cLowerLeft.fRed,cLowerRight.fRed,fractionS),
                   Math::lerp(cLowerLeft.fGreen,cLowerRight.fGreen,fractionS),
                   Math::lerp(cLowerLeft.fBlue,cLowerRight.fBlue,fractionS));

      flatTextureModulation.fRed = Math::lerp(cUpper.fRed,cLower.fRed,fractionT);
      flatTextureModulation.fGreen = Math::lerp(cUpper.fGreen,cLower.fGreen,fractionT);
      flatTextureModulation.fBlue = Math::lerp(cUpper.fBlue,cLower.fBlue,fractionT);
    }
    else {
      if (objectHit->fMaterial.fProceduralTextureType == ProceduralTexturer::kCheckerboard) {
        flatTextureModulation = fProceduralTexturer.sampleCheckerboard(s,t,
                                                                       objectHit->fMaterial.fProceduralTextureParameters[0],
                                                                       objectHit->fMaterial.fProceduralTextureParameters[1]);
      }
      else if (objectHit->fMaterial.fProceduralTextureType == ProceduralTexturer::kRadialColors) {
        flatTextureModulation = fProceduralTexturer.sampleRadialColors(s,t,
                                                                       Color(objectHit->fMaterial.fProceduralTextureParameters[0],
                                                                             objectHit->fMaterial.fProceduralTextureParameters[1],
                                                                             objectHit->fMaterial.fProceduralTextureParameters[2]),
                                                                       Color(objectHit->fMaterial.fProceduralTextureParameters[3],
                                                                             objectHit->fMaterial.fProceduralTextureParameters[4],
                                                                             objectHit->fMaterial.fProceduralTextureParameters[5]));
      }
    }
  }
  else {
    flatTextureEnabled = false;
  }

  // handle the ambient color by using the global ambient color of the scene
  Color ambientColor = fGlobalAmbientColor * objectHit->fMaterial.fAmbient;

  if (solidTextureEnabled) {
    // modulate the ambient color
    ambientColor *= solidTextureModulation;
  }

  if (flatTextureEnabled) {
    if (objectHit->fMaterial.fFlatTextureIsModulating) {
      // modulate the ambient color
      ambientColor *= flatTextureModulation;
    }
    else {
      // replace the ambient color
      ambientColor = flatTextureModulation;
    }
  }

  if (!objectHit->fMaterial.fRetainAmbientColor) {
    ambientColor *= (1.0 - objectHit->fMaterial.fTransparency);
  }

  color += ambientColor;

  if (flatTextureEnabled && (!objectHit->fMaterial.fFlatTextureIsModulating)) {
    // only ambient color is replaced
    // don't consider diffuse and specular components
  }
  else {
    // consider all possible light-contributions
    Lights::const_iterator light;
    for (light = fLights->begin(); light != fLights->end(); ++light) {

      // cast the shadowfeeler
      Vector3D s = Vector3D((*light)->fPosition) - Vector3D(hitPoint);
      shadowFeeler.fDirection = s;

      double shadowFactor = 1.0;

      if ((*light)->fIsSpotLight) {

        // find the angle between the shadowfeeler and the light direction
        Vector3D normalizedShadowFeeler(s);
        normalizedShadowFeeler.flip();
        normalizedShadowFeeler.normalize();

        Vector3D spotLightDirection((*light)->fSpotLightDirection);
        spotLightDirection.normalize();

        double cosAngle = normalizedShadowFeeler.inner(spotLightDirection);
        double angle = Math::radToDeg(acos(cosAngle));

        // check if the hitpoint lies inside the lightcone of the spotlight
        if ((angle <= (*light)->fCutOffAngle) &&
            (angle >= -(*light)->fCutOffAngle)) {

          // attenuate the light because of intersections with opaque/transparent geometric objects
          shadowFactor = calcShadowFactor(shadowFeeler);

          // attenuate the received light in the lightcone of the spotlight
          shadowFactor *= pow(cosAngle,(*light)->fExponent);
        }
        else {
          // the shadowfeeler doesn't lie inside the lightcone of the spotlight
          shadowFactor = 0.0;
        }
      }
      else {
        // attenuate the light because of intersections with opaque/transparent geometric objects
        shadowFactor = calcShadowFactor(shadowFeeler);
      }

      // don't handle ambient color per light (to prevent over-lighting)
      // it is already handled by using the scene's global ambient color

      Color lightContribution;

      // handle diffuse color
      s.normalize();
      double mDotS = s.inner(normal);
      if (objectHit->fMaterial.fLightBackFaces) {
        mDotS = Math::abs(mDotS);
      }
      if (mDotS > 0.0) {
        Color diffuse(objectHit->fMaterial.fDiffuse);
        // modulate diffuse component
        if (solidTextureEnabled) {
          diffuse *= solidTextureModulation;
        }
        if (flatTextureEnabled) {
          diffuse *= flatTextureModulation;
        }
        diffuse *= (*light)->fColor;
        diffuse *= mDotS;
        diffuse *= shadowFactor;
        if (!objectHit->fMaterial.fRetainDiffuseColor) {
          diffuse *= (1.0 - objectHit->fMaterial.fTransparency);
        }
        lightContribution += diffuse;
      }

      // handle specular reflection color
      Vector3D v(-ray.fDirection.fX,
                 -ray.fDirection.fY,
                 -ray.fDirection.fZ);
      v.normalize();
      Vector3D halfway(v.fX + s.fX,v.fY + s.fY,v.fZ + s.fZ);
      halfway.normalize();

      double mDotH = halfway.inner(normal);
      if (objectHit->fMaterial.fLightBackFaces) {
        mDotH = Math::abs(mDotH);
      }
      if (mDotH > 0.0) {
        double phong = pow(mDotH,objectHit->fMaterial.fSpecularExponent);
        Color specular(objectHit->fMaterial.fSpecular);
        specular *= (*light)->fColor;
        specular *= phong;
        specular *= shadowFactor;
        if (!objectHit->fMaterial.fRetainSpecularReflection) {
          specular *= (1.0 - objectHit->fMaterial.fTransparency);
        }
        lightContribution += specular;
      }

      // compute light source attenuation
      double lightSourceAttenuationFactor = 1.0;
      double distanceToLight = shadowFeeler.fDirection.length();
      double denominator = fLightAttenuationConstantFactor +
                           (fLightAttenuationLinearFactor * distanceToLight) +
                           (fLightAttenuationQuadraticFactor * distanceToLight * distanceToLight);
      if (Math::abs(denominator) > Math::kEpsilon) {
        denominator = 1.0 / denominator;
        if (denominator < 1.0) {
          lightSourceAttenuationFactor = denominator;
        }
      }

      color += (lightContribution * lightSourceAttenuationFactor);
    }
  }

  // blurred reflection (gloss) and blurred refraction (transparency) are
  // implemented using "non-weighted jittered discrete cone tracing"

  if (ray.fRecurseLevel < fMaxRecursionDepth) {

    bool followRay = ray.fFollowRay;
    Color rayColor = ray.fColor;

    if (ray.fRecurseLevel == 1) {
      followRay = (fFollowRays && (RNG::uniform() < fFollowRaysProbability));
      if (followRay) {
        // let the ray's color have a minimum brightness
        rayColor = Color(0.5 + RNG::uniform() / 2.0,
                         0.5 + RNG::uniform() / 2.0,
                         0.5 + RNG::uniform() / 2.0);
      }
    }

    // handle reflection (note that reflection inside an object is supported)
    if ((objectHit->fMaterial.fReflectivity >= fMinReflectivity) &&
        (objectHit->fMaterial.fReflectivity != 0.0)) {

      Ray reflectedRay(ray);
      reflectedRay.fFollowRay = followRay;
      reflectedRay.fColor = rayColor;

      // use a small offset to prevent self-reflection
      reflectedRay.fStart = Point3D(hitPoint.fX + (fReflectivityEpsilon * normal.fX),
                                    hitPoint.fY + (fReflectivityEpsilon * normal.fY),
                                    hitPoint.fZ + (fReflectivityEpsilon * normal.fZ));

      Vector3D direction = normalizedHitDirection;
      double v = 2.0 * direction.inner(normal);
      reflectedRay.fDirection = direction - (normal * v);
      reflectedRay.fDirection.normalize();

      ++reflectedRay.fRecurseLevel;
      reflectedRay.fIsReflective = true;
      reflectedRay.fIsRefractive = false;

      if ((objectHit->fMaterial.fGlossStrength != 0.0) &&
          (reflectedRay.fNrOfConeSamples != 1)) {

        vector<Ray> jitteredReflectedRays;
        Ray::jitterRaysInCone(jitteredReflectedRays,
                              reflectedRay,
                              objectHit->fMaterial.fGlossStrength / 2.0);

        for (int coneSample = 0; coneSample < reflectedRay.fNrOfConeSamples; ++coneSample) {
          Color computedColor(shootRay(jitteredReflectedRays[coneSample]) *
                              (objectHit->fMaterial.fReflectivity *
                               jitteredReflectedRays[coneSample].fSampleWeight));
          color += computedColor;
        }
      }
      else {
        color += Color(shootRay(reflectedRay) * objectHit->fMaterial.fReflectivity);
      }
    }

    // handle refraction (note that objects may interpenetrate each other)
    if ((objectHit->fMaterial.fTransparency >= fMinTransparency) &&
        (objectHit->fMaterial.fTransparency != 0.0)) {

      Ray refractedRay(ray);
      refractedRay.fStart = hitPoint;

      Vector3D direction(ray.fDirection);
      direction.normalize();

      refractedRay.fFollowRay = followRay;
      refractedRay.fColor = rayColor;

      // let c1 be the speed-of-light of the object with the highest priority
      // in the list of objects pierced by the ray
      double c1 = 1.0; // speed of light in air
      int maxPriority = 0;

      if (refractedRay.fNrOfObjectsPierced > 0) {      
        maxPriority = refractedRay.fObjectsPierced[0]->fMaterial.fPriority;
        c1 = refractedRay.fObjectsPierced[0]->fMaterial.fSpeedOfLight;
        for (int i = 1; i < refractedRay.fNrOfObjectsPierced; ++i) {
          if (refractedRay.fObjectsPierced[i]->fMaterial.fPriority > maxPriority) {
            maxPriority = refractedRay.fObjectsPierced[i]->fMaterial.fPriority;
            c1 = refractedRay.fObjectsPierced[i]->fMaterial.fSpeedOfLight;
          }
        }
      }

      double c2 = 1.0;
      double mDotDir;

      if (hitInfo.fIsEntering) {
        if (objectHit->fMaterial.fPriority > maxPriority) {
          c2 = objectHit->fMaterial.fSpeedOfLight;
        }
        else {
          c2 = c1;
        }

        // add the object to the pierced-objects list
        refractedRay.fObjectsPierced[refractedRay.fNrOfObjectsPierced] = objectHit;
        ++refractedRay.fNrOfObjectsPierced;

        mDotDir = normal.inner(direction);
        refractedRay.fIsRefractive = true;
      }
      else { // ray is exiting

        // find the current hit object in the pierced-objects list
        int objectPositionInList = 0;
        bool found = false;
        while ((objectPositionInList < refractedRay.fNrOfObjectsPierced) && (!found)) {
          if (refractedRay.fObjectsPierced[objectPositionInList] == objectHit) {
            found = true;
          }
          else {
            ++objectPositionInList;
          }
        }

        // remove the object from the pierced-objects list
        if (found) {
          for (int i = objectPositionInList; i < (refractedRay.fNrOfObjectsPierced - 1); ++i) {
            refractedRay.fObjectsPierced[i] = refractedRay.fObjectsPierced[i + 1];
          }
          --refractedRay.fNrOfObjectsPierced;
        }

        // let c2 be the speed-of-light of the object with the highest priority
        // in the list of objects pierced by the ray
        c2 = 1.0; // speed of light in air
        maxPriority = 0;
        if (refractedRay.fNrOfObjectsPierced > 0) {
          maxPriority = refractedRay.fObjectsPierced[0]->fMaterial.fPriority;
          c2 = refractedRay.fObjectsPierced[0]->fMaterial.fSpeedOfLight;
          for (int i = 1; i < refractedRay.fNrOfObjectsPierced; ++i) {
            if (refractedRay.fObjectsPierced[i]->fMaterial.fPriority > maxPriority) {
              maxPriority = refractedRay.fObjectsPierced[i]->fMaterial.fPriority;
              c2 = refractedRay.fObjectsPierced[i]->fMaterial.fSpeedOfLight;
            }
          }
        }

        normal.flip(); // invert normal because we're assumed to be inside the object !
        mDotDir = normal.inner(direction);

        if (refractedRay.fNrOfObjectsPierced == 0) {
          refractedRay.fIsRefractive = false;
        }
      }

      double c2Divc1 = c2 / c1;
      double sqrtArg = 1.0 - ((c2Divc1 * c2Divc1) * (1.0 - (mDotDir * mDotDir)));

      // don't continue with a negative argument for the square-root
      // (i.e.: check for total internal reflection when beyond the critical angle)
      if ((sqrtArg >= 0.0) ||
          objectHit->fMaterial.fDisableRefraction) {

        if (!objectHit->fMaterial.fDisableRefraction) {
          double cosTheta2 = sqrt(sqrtArg);
          refractedRay.fDirection = (direction * c2Divc1) +
                                    (normal * ((mDotDir * c2Divc1) - cosTheta2));
        }
        refractedRay.fDirection.normalize();

        // use a small offset to prevent self-transparency
        refractedRay.fStart = Point3D(refractedRay.fStart.fX + (fTransparencyEpsilon * refractedRay.fDirection.fX),
                                      refractedRay.fStart.fY + (fTransparencyEpsilon * refractedRay.fDirection.fY),
                                      refractedRay.fStart.fZ + (fTransparencyEpsilon * refractedRay.fDirection.fZ));

        ++refractedRay.fRecurseLevel;
        refractedRay.fIsReflective = false;

        if ((objectHit->fMaterial.fTranslucencyStrength != 0.0) &&
            (refractedRay.fNrOfConeSamples != 1)) {

          vector<Ray> jitteredRefractedRays;
          Ray::jitterRaysInCone(jitteredRefractedRays,
                                refractedRay,
                                objectHit->fMaterial.fTranslucencyStrength / 2.0);

          for (int coneSample = 0; coneSample < refractedRay.fNrOfConeSamples; ++coneSample) {

            Color computedColor;
            computedColor = shootRay(jitteredRefractedRays[coneSample]) *
                            (objectHit->fMaterial.fTransparency *
                             jitteredRefractedRays[coneSample].fSampleWeight);

            color += computedColor;

          }
        }
        else {
          color += Color(shootRay(refractedRay) * objectHit->fMaterial.fTransparency);
        }
      }
    }
  }

  // compute atmospheric attenuation (depth cueing) based on the real distance (not only the z-coordinate)
  double distanceToHitPoint = Vector3D(Vector3D(hitPoint) - Vector3D(fEyePoint)).length();
  double scale = fAtmosphericAttenuationFrontScale;

  if (distanceToHitPoint <= fAtmosphericAttenuationFrontPlane) {
    // don't change scale
  }
  else if (distanceToHitPoint >= fAtmosphericAttenuationBackPlane) {
    scale = fAtmosphericAttenuationBackScale;
  }
  else if (fAtmosphericAttenuationFrontPlane != fAtmosphericAttenuationBackPlane) {
    // linearly interpolate between the front scale and back scale
    scale = fAtmosphericAttenuationBackScale +
            (((distanceToHitPoint - fAtmosphericAttenuationBackPlane) *
              (fAtmosphericAttenuationFrontScale - fAtmosphericAttenuationBackScale)) /
             (fAtmosphericAttenuationFrontPlane - fAtmosphericAttenuationBackPlane));
  }

  // blend in the atmospheric attenuation color
  color = (color * scale) + (fAtmosphericAttenuationColor * (1.0 - scale));

  // compute the fog based on the real distance (not only the z-coordinate)
  double fogFactor = 1.0;

  if (fFogStart != fFogEnd) {
    if (distanceToHitPoint >= fFogEnd) {
      fogFactor = 0.0;
    }
    else if (distanceToHitPoint <= fFogStart) {
      fogFactor = 1.0;
    }
    else {
      fogFactor = (fFogEnd - distanceToHitPoint) / (fFogEnd - fFogStart);
    }
  }

  // blend in the fog
  color = (color * fogFactor) + (fFogColor * (1.0 - fogFactor));

  return color;
}

void Scene::computeShadowsWhenRayTracing(void)
{
  fComputeShadows = true;
}

void Scene::dontComputeShadowsWhenRayTracing(void)
{
  fComputeShadows = false;
}

void Scene::computeSoftShadowsWhenRayTracing(void)
{
  fComputeSoftShadows = true;
}

void Scene::computeHardShadowsWhenRayTracing(void)
{
  fComputeSoftShadows = false;
}

void Scene::setShadowFeelerJitterStrength(double shadowFeelerJitterStrength)
{
  fShadowFeelerJitterStrength = shadowFeelerJitterStrength;
}

void Scene::findNormalsWhenRayTracing(double findNormalsProbability)
{
  fFindNormals = true;
  fFindNormalsProbability = findNormalsProbability;
}

void Scene::dontFindNormalsWhenRayTracing(void)
{
  fFindNormals = false;
}

void Scene::followRaysWhenRayTracing(double followRaysProbability)
{
  fFollowRays = true;
  fFollowRaysProbability = followRaysProbability;
}

void Scene::dontFollowRaysWhenRayTracing(void)
{
  fFollowRays = false;
}

bool Scene::isMotionBlurEnabled(void) const
{
  return fMotionBlurEnabled;
}

double Scene::getMotionBlurFrameLength(void) const
{
  return fMotionBlurFrameLength;
}

int Scene::getNrOfMotionBlurTimeSamples(void) const
{
  return fNrOfMotionBlurTimeSamples;
}

void Scene::applyShapesMotionBlurAffineTransformation(double timeIndex)
{
  GeometricObjects::const_iterator p;

  for (p = fGeometricObjects->begin(); p != fGeometricObjects->end(); ++p) {
    (*p)->applyMotionBlurTransformation(timeIndex,fMotionBlurFrameLength);
  }  
}

void Scene::restoreShapes(void)
{
  GeometricObjects::const_iterator p;

  for (p = fGeometricObjects->begin(); p != fGeometricObjects->end(); ++p) {
    (*p)->restoreOriginalTransformation();
  }  
}

long Scene::getNrOfRaysShot(void) const
{
  return fNrOfRaysShot;
}

void Scene::resetStatistics(void)
{
  fNrOfRaysShot = 0;
}

// ---------------
// utility methods
// ---------------
void Scene::initialize(void)
{
  fSceneLoaded = false;
  fDefStack = 0;
  fLights = 0;
  fGeometricObjects = 0;
  fEyePointSpecified = false;
  fViewPointSpecified = false;
  fViewAngleSpecified = false;
  fUpDirectionSpecified = false;
  fEyePoint.set(0.0,0.0,1.0);
  fViewPoint.set(0.0,0.0,0.0);
  fViewAngle = 45.0;
  fUpDirection.set(0.0,1.0,0.0);
  fLines.clear();
  fBackgroundColor = Color(0.0,0.0,0.3);
  fGlobalAmbientColor = Color(0.1,0.1,0.1);
  fLightAttenuationConstantFactor = 1.0;
  fLightAttenuationLinearFactor = 0.0;
  fLightAttenuationQuadraticFactor = 0.0;
  fAtmosphericAttenuationFrontPlane = 0.0;
  fAtmosphericAttenuationFrontScale = 1.0;
  fAtmosphericAttenuationBackPlane = 0.0;
  fAtmosphericAttenuationBackScale = 1.0;
  fAtmosphericAttenuationColor = Color(0.0,0.0,0.0);
  fFogStart = 0.0;
  fFogEnd = 0.0;
  fFogColor = Color(0.0,0.0,0.0);
  fShadowFeelerEpsilon = 1E-6;
  fReflectivityEpsilon = 1E-6;
  fTransparencyEpsilon = 1E-6;
  fAdaptedShadowComputation = false;
  fMinReflectivity = 0.1;
  fMinTransparency = 0.1;
  fMaxRecursionDepth = 5;
  computeShadowsWhenRayTracing();
  computeHardShadowsWhenRayTracing();
  setShadowFeelerJitterStrength(0.05);
  fShowExtents = false;
  fLights = new Lights();
  fGeometricObjects = new GeometricObjects();
  fCurrentMaterial.setDefault();
  fSolidTexturer.initialize();
  fProceduralTexturer.initialize();
  dontFindNormalsWhenRayTracing();
  fMotionBlurEnabled = false;
  fMotionBlurFrameLength = 1.0;
  fNrOfMotionBlurTimeSamples = 8;
  resetMotionBlurTransformations();
  resetStatistics();
}

GeometricObject* Scene::getGeometricObject(void)
{
  string keyword, name;
  GeometricObject* shape;
  ETokenType tokenType;
  Point3D point;
  Color color;
  Light* light = 0;
  double angle;
  Vector3D v;
  double sx, sy, sz;
  Vector3D row1, row2, row3;
  stringstream* tempInputStream;
  string filename, leftBrace, temp, inputstr;
  int i, j;
  Scene* includedScene;
  Lights::const_iterator l;
  GeometricObjects::const_iterator p;

  while ((tokenType = (getToken(keyword = getNextToken()))) != kT_NULL) {

    if ((tokenType == kUNIONBOOL) ||
        (tokenType == kINTERSECTIONBOOL) ||
        (tokenType == kDIFFERENCEBOOL)) {
      switch(tokenType) {

        case kUNIONBOOL :
          shape = new UnionBool();
          break;

        case kINTERSECTIONBOOL :
          shape = new IntersectionBool();
          break;

        case kDIFFERENCEBOOL :
          shape = new DifferenceBool();
          break;
      }

      GeometricObject* p = shape;

      // get left child
      p = getGeometricObject();
      if (p == 0) {
         // there should always be a left object
         return 0;
      }
      ((Boolean*) shape)->fLeft = p;

      // get right child
      p = getGeometricObject();
      if (p == 0) {
         // there should always be a right object
        return 0;
      }
      ((Boolean*) shape)->fRight = p;

      return shape;
    }
    else {
      switch(tokenType) {

        case kINCLUDE :
          filename = getNextToken();
          if (getToken(filename) != kIDENT) {
            cerr << "Scene::getGeometricObject() -> filename expected at line #"
                 << fCurrentLine << "." << endl;
            exit(0);
          }
          else {
            // load and parse the file
            includedScene = new Scene(filename);

            if (includedScene->sceneLoaded()) {

              // merge the definition-stacks
              fDefStack->add(includedScene->fDefStack);

              // merge the lights of the included scene
              for (l = includedScene->fLights->begin();
                l != includedScene->fLights->end(); ++l) {
                addLight(*l);
              }

              // merge the geometric objects of the included scene
              for (p = includedScene->fGeometricObjects->begin();
                p != includedScene->fGeometricObjects->end(); ++p) {
                addGeometricObject(*p);
              }

            }

            delete includedScene;
          }
          break;

        case kLIGHT :
          light = new Light();
          light->fPosition.fX = getDouble();
          light->fPosition.fY = getDouble();
          light->fPosition.fZ = getDouble();
          light->fColor.fRed = getDouble();
          light->fColor.fGreen = getDouble();
          light->fColor.fBlue = getDouble();
          light->fIsSpotLight = false;
          fLights->push_back(light);
          break;

        case kSPOTLIGHT :
          light = new Light();
          light->fPosition.fX = getDouble();
          light->fPosition.fY = getDouble();
          light->fPosition.fZ = getDouble();
          light->fColor.fRed = getDouble();
          light->fColor.fGreen = getDouble();
          light->fColor.fBlue = getDouble();
          light->fIsSpotLight = true;
          light->fSpotLightDirection.fX = getDouble();
          light->fSpotLightDirection.fY = getDouble();
          light->fSpotLightDirection.fZ = getDouble();
          light->fCutOffAngle = getDouble();
          light->fExponent = getDouble();
          fLights->push_back(light);
          break;

        case kIDENTITYAFFINE :
          fAffineStack.setIdentity();
          break;

        case kROTATE :
          angle = getDouble();
          v.fX = getDouble();
          v.fY = getDouble();
          v.fZ = getDouble();
          fAffineStack.rotate(angle,v);
          break;

        case kTRANSLATE :
          v.fX = getDouble();
          v.fY = getDouble();
          v.fZ = getDouble();
          fAffineStack.translate(v);
          break;

        case kSCALE :
          sx = getDouble();
          sy = getDouble();
          sz = getDouble();
          fAffineStack.scale(sx,sy,sz);
          break;

        case kTRANSFORM :
          row1.fX = getDouble();
          row1.fY = getDouble();
          row1.fZ = getDouble();
          row2.fX = getDouble();
          row2.fY = getDouble();
          row2.fZ = getDouble();
          row3.fX = getDouble();
          row3.fY = getDouble();
          row3.fZ = getDouble();
          fAffineStack.transform(row1,row2,row3);
          break;

        case kPUSH :
          fAffineStack.duplicate();
          break; 

        case kPOP :
          fAffineStack.pop();
          break;

        case kMBROTATE :
          fMotionBlurRotationAngle = getDouble();
          fMotionBlurRotationAxis.fX = getDouble();
          fMotionBlurRotationAxis.fY = getDouble();
          fMotionBlurRotationAxis.fZ = getDouble();
          break;

        case kMBTRANSLATE :
          fMotionBlurTranslation.fX = getDouble();
          fMotionBlurTranslation.fY = getDouble();
          fMotionBlurTranslation.fZ = getDouble();
          break;

        case kMBSCALE :
          fMotionBlurScale.fX = getDouble();
          fMotionBlurScale.fY = getDouble();
          fMotionBlurScale.fZ = getDouble();
          if ((Math::abs(fMotionBlurScale.fX) < Math::kEpsilon) ||
              (Math::abs(fMotionBlurScale.fY) < Math::kEpsilon) ||
              (Math::abs(fMotionBlurScale.fZ) < Math::kEpsilon)) {
            // degenerate scaling transformation
            fMotionBlurScale = Vector3D(1.0,1.0,1.0);
          }
          break;
      
        case kMBCLEAR :
          resetMotionBlurTransformations();
          break;

        case kMBENABLE :
          fCurrentMaterial.fMotionBlurEnabled = true;
          break;

        case kMOTIONBLUR :
          fMotionBlurEnabled = true;
          fMotionBlurFrameLength = getDouble();
          fNrOfMotionBlurTimeSamples = (int) getDouble();
          if (fNrOfMotionBlurTimeSamples < 2) {
            fNrOfMotionBlurTimeSamples = 2;
          }
          break;

        case kSURFACEROUGHNESS :
          fCurrentMaterial.fSurfaceRoughness = getDouble();
          break;

        case kEMISSIVE :
          color.fRed = getDouble();
          color.fGreen = getDouble();
          color.fBlue = getDouble();
          fCurrentMaterial.fEmissive = color;
          break;

        case kAMBIENT :
          color.fRed = getDouble();
          color.fGreen = getDouble();
          color.fBlue = getDouble();
          fCurrentMaterial.fAmbient = color;
          break;

        case kDIFFUSE :
          color.fRed = getDouble();
          color.fGreen = getDouble();
          color.fBlue = getDouble();
          fCurrentMaterial.fDiffuse = color;
          break;

        case kSPECULAR :
          color.fRed = getDouble();
          color.fGreen = getDouble();
          color.fBlue = getDouble();
          fCurrentMaterial.fSpecular = color;
          break;

        case kSPECULAREXPONENT :
          fCurrentMaterial.fSpecularExponent = getDouble();
          if (fCurrentMaterial.fSpecularExponent < 0.0) {
            fCurrentMaterial.fSpecularExponent = 0.0;
          }
          else if (fCurrentMaterial.fSpecularExponent > 128.0) {
            fCurrentMaterial.fSpecularExponent = 128.0;
          }
          break;

        case kLIGHTBACKFACES :
          fCurrentMaterial.fLightBackFaces = (getDouble() != 0.0);
          break;

        case kDEFAULTMATERIALS :
          fCurrentMaterial.setDefault();
          break;

        case kSPEEDOFLIGHT :
          fCurrentMaterial.fSpeedOfLight = getDouble();
          break;

        case kTRANSPARENCY :
          fCurrentMaterial.fTransparency = getDouble();
          break;

        case kREFLECTIVITY :
          fCurrentMaterial.fReflectivity = getDouble();
          break;

        case kGLOSSSTRENGTH :
          fCurrentMaterial.fGlossStrength = getDouble();
          break;

        case kTRANSLUCENCYSTRENGTH :
          fCurrentMaterial.fTranslucencyStrength = getDouble();
          break;

        case kPRIORITY :
          fCurrentMaterial.fPriority = (int) getDouble();
          break;

        case kRETAINAMBIENTCOLOR :
          fCurrentMaterial.fRetainAmbientColor = (getDouble() != 0.0);
          break;

        case kRETAINDIFFUSECOLOR :
          fCurrentMaterial.fRetainDiffuseColor = (getDouble() != 0.0);
          break;

        case kRETAINSPECULARREFLECTION :
          fCurrentMaterial.fRetainSpecularReflection = (getDouble() != 0.0);
          break;

        case kDISABLEREFRACTION :
          fCurrentMaterial.fDisableRefraction = (getDouble() != 0.0);
          break;

        case kEYEPOINT :
          fEyePoint.fX = getDouble();
          fEyePoint.fY = getDouble();
          fEyePoint.fZ = getDouble();
          fEyePointSpecified = true;
          break;

        case kVIEWPOINT :
          fViewPoint.fX = getDouble();
          fViewPoint.fY = getDouble();
          fViewPoint.fZ = getDouble();
          fViewPointSpecified = true;
          break;

        case kVIEWANGLE :
          fViewAngle = getDouble();
          fViewAngleSpecified = true;
          break;

        case kUPDIRECTION :
          fUpDirection.fX = getDouble();
          fUpDirection.fY = getDouble();
          fUpDirection.fZ = getDouble();
          fUpDirectionSpecified = true;
          break;

        case kBACKGROUND :
          fBackgroundColor.fRed = getDouble();
          fBackgroundColor.fGreen = getDouble();
          fBackgroundColor.fBlue = getDouble();
          break;

        case kGLOBALAMBIENT :
          fGlobalAmbientColor.fRed = getDouble();
          fGlobalAmbientColor.fGreen = getDouble();
          fGlobalAmbientColor.fBlue = getDouble();
          break;

        case kLIGHTATTENUATION :
          fLightAttenuationConstantFactor = getDouble();
          fLightAttenuationLinearFactor = getDouble();
          fLightAttenuationQuadraticFactor = getDouble();
          break;

        case kATMOSPHERICATTENUATION :
          fAtmosphericAttenuationFrontPlane = getDouble();
          fAtmosphericAttenuationFrontScale = getDouble();
          fAtmosphericAttenuationBackPlane = getDouble();
          fAtmosphericAttenuationBackScale = getDouble();
          fAtmosphericAttenuationColor.fRed = getDouble();
          fAtmosphericAttenuationColor.fGreen = getDouble();
          fAtmosphericAttenuationColor.fBlue = getDouble();

          // clamp the values
          if (fAtmosphericAttenuationFrontScale < 0.0) {
            fAtmosphericAttenuationFrontScale = 0.0;
          }
          else if (fAtmosphericAttenuationFrontScale > 1.0) {
            fAtmosphericAttenuationFrontScale = 1.0;
          }
          if (fAtmosphericAttenuationBackScale < 0.0) {
            fAtmosphericAttenuationBackScale = 0.0;
          }
          else if (fAtmosphericAttenuationBackScale > 1.0) {
            fAtmosphericAttenuationBackScale = 1.0;
          }
          break;

        case kFOG :
          fFogStart = getDouble();
          fFogEnd = getDouble();
          fFogColor.fRed = getDouble();
          fFogColor.fGreen = getDouble();
          fFogColor.fBlue = getDouble();
          break;

        case kSHADOWFEELEREPSILON :
          fShadowFeelerEpsilon = getDouble();
          break;

        case kREFLECTIVITYEPSILON :
          fReflectivityEpsilon = getDouble();
          break;

        case kTRANSPARENCYEPSILON :
          fTransparencyEpsilon = getDouble();
          break;

        case kMESHEPSILON :
          fMeshEpsilon = getDouble();
          break;

        case kADAPTEDSHADOWCOMPUTATION :
          fAdaptedShadowComputation = (getDouble() != 0.0);
          break;

        case kMINREFLECTIVITY :
          fMinReflectivity = getDouble();
          break;

        case kMINTRANSPARENCY :
          fMinTransparency = getDouble();
          break;

        case kMAXRECURSIONDEPTH :
          fMaxRecursionDepth = (int) getDouble();
          break;

        case kSHOWEXTENTS :
          fShowExtents = (getDouble() != 0.0);
          break;

        case kSOLIDTEXTURE :
          i = (int) getDouble();
          switch (i) {

            case 1 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kCheckerboard;
              for (j = 0; j < 3; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 2 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kStackedColorBars;
              for (j = 0; j < 3; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 3 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kRGBCubes;
              for (j = 0; j < 3; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 4 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kOneCoordinateBasedContourLines;
              for (j = 0; j < 2; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 5 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kTwoCoordinatesBasedContourLines;
              for (j = 0; j < 4; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 6 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kThreeCoordinatesBasedContourLines;
              for (j = 0; j < 4; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 7 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kWoodGrain;
              for (j = 0; j < 11; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 8 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kMarble;
              for (j = 0; j < 5; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            case 9 :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kCowSpots;
              for (j = 0; j < 1; ++j) {
                fCurrentMaterial.fSolidTextureParameters[j] = getDouble();
              }
              break;

            default :
              fCurrentMaterial.fSolidTextureType = SolidTexturer::kNone;
              break;
          }
          break;

        case kIMAGETEXTURE :
          filename = getNextToken();
          if (getToken(filename) != kIDENT) {
            cerr << "Scene::getGeometricObject() -> filename expected at line #"
                 << fCurrentLine << "." << endl;
            exit(0);
          }
          else {
            fCurrentMaterial.fHasImageTexture = true;
            fCurrentMaterial.fImageTextureBitmap = new RGBApixmap();
            if (!fCurrentMaterial.fImageTextureBitmap->loadFromBMPFile(filename)) {
              exit(0);
            }
          }
          fCurrentMaterial.fFlatTextureIsModulating = (getDouble() != 0.0);
          fCurrentMaterial.fImageTextureTileS = getDouble();
          fCurrentMaterial.fImageTextureTileT = getDouble();
          break;

        case kPROCEDURALTEXTURE :
          i = (int) getDouble();
          fCurrentMaterial.fFlatTextureIsModulating = (getDouble() != 0.0);
          switch (i) {

            case 1 :
              fCurrentMaterial.fProceduralTextureType = ProceduralTexturer::kCheckerboard;
              for (j = 0; j < 2; ++j) {
                fCurrentMaterial.fProceduralTextureParameters[j] = getDouble();
              }
              break;

            case 2 :
              fCurrentMaterial.fProceduralTextureType = ProceduralTexturer::kRadialColors;
              for (j = 0; j < 6; ++j) {
                fCurrentMaterial.fProceduralTextureParameters[j] = getDouble();
              }
              break;

            default :
              fCurrentMaterial.fProceduralTextureType = ProceduralTexturer::kNone;
              break;
          }
          break;

        case kFLATTEXTUREOFFSET :
            fCurrentMaterial.fFlatTextureSOffset = getDouble();
            fCurrentMaterial.fFlatTextureTOffset = getDouble();
          break;

        case kT_NULL :
          break; // the null-token represents end-of-file 

        case kDEF :
          name = getNextToken();
          if (getToken(name) != kIDENT) {
            cerr << "Scene::getGeometricObject() -> identifier expected at line #"
                 << fCurrentLine << "." << endl;
            return 0;
          }

          if (fDefStack->search(name)) {
            cerr << "Scene::getGeometricObject() -> " << name << " redefined at line #"
                 << fCurrentLine << "." << endl;
            return 0;
          }

          leftBrace = getNextToken();
          if (getToken(leftBrace) != kLEFTBRACE) {
            cerr << "Scene::getGeometricObject() -> { expected at line #"
                 << fCurrentLine << "." << endl;
            return 0;
          }

          temp = "";
          inputstr = "";
          while (getToken(temp = getNextToken()) != kRIGHTBRACE) {
            inputstr += temp + " ";
            if (!fInputStream) {
              cerr << "Scene::getGeometricObject() -> end of file detected at line #"
                   << fCurrentLine << "." << endl;
              return 0;
            }
          }

          fDefStack->push(name,inputstr);
          break;

        case kUSE :
          name = getNextToken();
          if (getToken(name) != kIDENT) {
            cerr << "Scene::getGeometricObject() -> identifier expected at line #"
                 << fCurrentLine << "." << endl;
            return 0;
          }

          if (!fDefStack->search(name)) {
            cerr << "Scene::getGeometricObject() -> " << name
                 << " not defined at line #" << fCurrentLine << "." << endl;
            return 0;
          }

          tempInputStream = new stringstream;
          *tempInputStream << fDefStack->contents(name) << " ";
          *tempInputStream << fInputStream->rdbuf();
          delete (fInputStream);
          fInputStream = tempInputStream;
          break;

        default :
          // handle all the shapes
          switch(tokenType) {

            case kBUCKYBALL :
              shape = new Mesh("data/meshes/3vn/buckyball.3vn",Mesh::k3VN);
              break;

            case kCONE :
              shape = new TaperedCylinder();
              ((TaperedCylinder*) shape)->fSmallRadius = 0.0;
              ((TaperedCylinder*) shape)->fDistortTexture = (getDouble() != 0.0);
              break;

            case kCUBE :
              shape = new Cube();
              break;

            case kCYLINDER :
              shape = new TaperedCylinder();
              ((TaperedCylinder*) shape)->fSmallRadius = 1.0;
              ((TaperedCylinder*) shape)->fDistortTexture = (getDouble() != 0.0);
              break;

            case kDIAMOND :
              shape = new Mesh("data/meshes/3vn/diamond.3vn",Mesh::k3VN);
              break;

            case kDODECAHEDRON :
              shape = new Mesh("data/meshes/3vn/dodecahedron.3vn",Mesh::k3VN);
              break;

            case kICOSAHEDRON :
              shape = new Mesh("data/meshes/3vn/icosahedron.3vn",Mesh::k3VN);
              break;

            case kMESH :
              filename = getNextToken();
              // retrieve the type of mesh by examining the file's extension
              i = filename.length();
              temp = filename.substr(i - 3,3);
              if ((temp == "3vn") ||
                  (temp == "3VN")) {
                shape = new Mesh(filename,Mesh::k3VN);
              }
              else if ((temp == "3dv") ||
                       (temp == "3DV")) {
                shape = new Mesh(filename,Mesh::k3DV);
              }
              else if ((temp == "3dm") ||
                       (temp == "3DM")) {
                shape = new Mesh(filename,Mesh::k3DM);
              }
              else if ((temp == "obj") ||
                       (temp == "OBJ")) {
                shape = new Mesh(filename,Mesh::kOBJ);
              }
              ((Mesh*) shape)->triangulateFaces();
              ((Mesh*) shape)->fInterpolateNormals = (getDouble() != 0.0);
              break;

            case kOCTAHEDRON :
              shape = new Mesh("data/meshes/3vn/octahedron.3vn",Mesh::k3VN);
              break;

            case kPYRAMID :
              shape = new Mesh("data/meshes/3dm/pyramid.3dm",Mesh::k3DM);
              ((Mesh*) shape)->triangulateFaces();
              break;

            case kSPHERE :
              shape = new Sphere();
              break;

            case kTAPEREDCYLINDER :
              shape = new TaperedCylinder();
              ((TaperedCylinder*) shape)->fSmallRadius = getDouble();
              ((TaperedCylinder*) shape)->fDistortTexture = (getDouble() != 0.0);
              break;

            case kTEAPOT :
              shape = new Mesh("data/meshes/obj/teapot.obj",Mesh::kOBJ);
              ((Mesh*) shape)->triangulateFaces();
              ((Mesh*) shape)->fInterpolateNormals = true;
              break;

            case kTETRAHEDRON :
              shape = new Mesh("data/meshes/3vn/tetrahedron.3vn",Mesh::k3VN);
              break;

            case kTORUS :
              shape = new Torus();
              ((Torus*) shape)->fTubeRadius = getDouble();
              ((Torus*) shape)->fTorusRadius = getDouble();
              break;

            case kONESHEETHYPERBOLOID :
              shape = new OneSheetHyperboloid();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kTWOSHEETHYPERBOLOID :
              shape = new TwoSheetHyperboloid();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kELLIPTICCYLINDER :
              shape = new EllipticCylinder();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kELLIPTICCONE :
              shape = new EllipticCone();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kELLIPTICPARABOLOID :
              shape = new EllipticParaboloid();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kHYPERBOLICPARABOLOID :
              shape = new HyperbolicParaboloid();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kPINCHEDSURFACEWITHHOLES :
              shape = new PinchedSurfaceWithHoles();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kPINCHEDSURFACEWITHLOBE :
              shape = new PinchedSurfaceWithLobe();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kFOURLOBES :
              shape = new FourLobes();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kSTEINERSROMANSURFACE :
              shape = new SteinersRomanSurface();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            case kTOOTHCUBE :
              shape = new ToothCube();
              ((AlgebraicSurface*) shape)->fBoxExtentWidth = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentHeight = getDouble();
              ((AlgebraicSurface*) shape)->fBoxExtentDepth = getDouble();
              break;

            default :
              cerr << "Scene::getGeometricObject() -> "
                   << "unknown keyword [" << keyword
                   << "] at line #" << fCurrentLine << "." << endl;
              return 0;
          }

          ((Shape*) shape)->fMaterial.set(fCurrentMaterial);
          ((Shape*) shape)->fOriginalAffineTransformation.set(*(fAffineStack.fTopOfStack->fAffineTransformation));
          ((Shape*) shape)->fOriginalInverseAffineTransformation.set(*(fAffineStack.fTopOfStack->fInverseAffineTransformation));
          ((Shape*) shape)->fAffineTransformation.set(*(fAffineStack.fTopOfStack->fAffineTransformation));
          ((Shape*) shape)->fInverseAffineTransformation.set(*(fAffineStack.fTopOfStack->fInverseAffineTransformation));
          ((Shape*) shape)->fMotionBlurRotationAngle = fMotionBlurRotationAngle;
          ((Shape*) shape)->fMotionBlurRotationAxis = fMotionBlurRotationAxis;
          ((Shape*) shape)->fMotionBlurTranslation = fMotionBlurTranslation;
          ((Shape*) shape)->fMotionBlurScale = fMotionBlurScale;
          ((Shape*) shape)->computeExtent();

          resetMotionBlurTransformations();

          return shape;
      } // switch
    } // else
  } // while

  return 0;
}

string Scene::getNextToken(void)
{
  char ch;
  string token;
  int lastChar = 1;

  if (!fInputStream) {
    return token;
  }

  if (fInputStream->eof()) {
    return token;
  }

  while (fInputStream->get(ch)) {
    if (fInputStream->eof()) {
      return token;
    }

    switch (ch) {
      case '\n' :
        ++fNextLine;

      case ' '  :
      case '\t' :
      case '\a' :
      case '\b' :
      case '\v' :
      case '\f' :
      case '\r' :
        if (lastChar == 0) {
          return token;
        }
        break;

      case '{' :
        token = ch;
        return token;
        break;

      case '}' :
        token = ch;
        return token;
        break;

      case '!' :
        while ((ch != '\n') &&
               (fInputStream->get(ch))) {
          // ignore comments
        }
        ++fNextLine;
        break;

      default :
        token += ch;
        lastChar = 0;
        if ((fInputStream->peek() == '{') ||
            (fInputStream->peek() == '}')) {
        if (lastChar == 0) {
          return token;
        }
        else {
          fInputStream->get(ch);
          token = ch;
          return token;
        }
      }
      fCurrentLine = fNextLine;
    }
  }

  return " ";
}

double Scene::getDouble(void)
{
  stringstream tmp;
  double number;
  string str = getNextToken();

  tmp << str;
  if (!(tmp >> number)) {
    cerr << "Scene::getDouble() -> error getting double at line #" << fCurrentLine << "." << endl;
    exit(1);
  }
  else {
    char t;
    if (tmp >> t) {
      cerr << "Scene::getDouble() -> bum chars in double at line #" << fCurrentLine << "." << endl;
      exit(1);
    } 
  }

  return number;
}

bool Scene::isIdentifier(string keyword) const
{
  string temp = keyword;

  if (!isalpha(temp[0])) {
    return false;
  }

  for (int count = 1; count < temp.length(); ++count) {
    if ((!isalnum(temp[count])) &&
        (temp[count] != '.') &&
        (temp[count] != '/') &&
        (temp[count] != '-')) {
      return false;
    }
  }

  return true;
}

Scene::ETokenType Scene::getToken(string keyword) const
{
  string temp = keyword;

  // convert keyword/identifier to it's uppercase equivalent
  for (int i = 0; i < temp.length(); ++i) {
    if ((((int) temp[i]) >= 97) &&
        (((int) temp[i]) <= 122)) {
      temp[i] = toupper(temp[i]);
    }
  }

       if (temp == "LIGHT")                    { return kLIGHT;                    }
  else if (temp == "INCLUDE")                  { return kINCLUDE;                  }
  else if (temp == "SPOTLIGHT")                { return kSPOTLIGHT;                }
  else if (temp == "IDENTITYAFFINE")           { return kIDENTITYAFFINE;           }
  else if (temp == "ROTATE")                   { return kROTATE;                   }
  else if (temp == "TRANSLATE")                { return kTRANSLATE;                }
  else if (temp == "SCALE")                    { return kSCALE;                    }
  else if (temp == "TRANSFORM")                { return kTRANSFORM;                }
  else if (temp == "PUSH")                     { return kPUSH;                     }
  else if (temp == "POP")                      { return kPOP;                      }
  else if (temp == "MBCLEAR")                  { return kMBCLEAR;                  }
  else if (temp == "MBROTATE")                 { return kMBROTATE;                 }
  else if (temp == "MBTRANSLATE")              { return kMBTRANSLATE;              }
  else if (temp == "MBSCALE")                  { return kMBSCALE;                  }
  else if (temp == "MBENABLE")                 { return kMBENABLE;                 }
  else if (temp == "MOTIONBLUR")               { return kMOTIONBLUR;               }
  else if (temp == "BUCKYBALL")                { return kBUCKYBALL;                }
  else if (temp == "CONE")                     { return kCONE;                     }
  else if (temp == "CUBE")                     { return kCUBE;                     }
  else if (temp == "CYLINDER")                 { return kCYLINDER;                 }
  else if (temp == "DIAMOND")                  { return kDIAMOND;                  }
  else if (temp == "DODECAHEDRON")             { return kDODECAHEDRON;             }
  else if (temp == "ICOSAHEDRON")              { return kICOSAHEDRON;              }
  else if (temp == "MESH")                     { return kMESH;                     }
  else if (temp == "OCTAHEDRON")               { return kOCTAHEDRON;               }
  else if (temp == "PYRAMID")                  { return kPYRAMID;                  }
  else if (temp == "SPHERE")                   { return kSPHERE;                   }
  else if (temp == "TAPEREDCYLINDER")          { return kTAPEREDCYLINDER;          }
  else if (temp == "TEAPOT")                   { return kTEAPOT;                   }
  else if (temp == "TETRAHEDRON")              { return kTETRAHEDRON;              }
  else if (temp == "TORUS")                    { return kTORUS;                    }
  else if (temp == "ONESHEETHYPERBOLOID")      { return kONESHEETHYPERBOLOID;      }
  else if (temp == "TWOSHEETHYPERBOLOID")      { return kTWOSHEETHYPERBOLOID;      }
  else if (temp == "ELLIPTICCYLINDER")         { return kELLIPTICCYLINDER;         }
  else if (temp == "ELLIPTICCONE")             { return kELLIPTICCONE;             }
  else if (temp == "ELLIPTICPARABOLOID")       { return kELLIPTICPARABOLOID;       }
  else if (temp == "HYPERBOLICPARABOLOID")     { return kHYPERBOLICPARABOLOID;     }
  else if (temp == "PINCHEDSURFACEWITHHOLES")  { return kPINCHEDSURFACEWITHHOLES;  }
  else if (temp == "PINCHEDSURFACEWITHLOBE")   { return kPINCHEDSURFACEWITHLOBE;   }
  else if (temp == "FOURLOBES")                { return kFOURLOBES;                }
  else if (temp == "STEINERSROMANSURFACE")     { return kSTEINERSROMANSURFACE;     }
  else if (temp == "TOOTHCUBE")                { return kTOOTHCUBE;                }
  else if (temp == "DIFFERENCE")               { return kDIFFERENCEBOOL;           }
  else if (temp == "INTERSECTION")             { return kINTERSECTIONBOOL;         }
  else if (temp == "UNION")                    { return kUNIONBOOL;                }
  else if (temp == "DEFAULTMATERIALS")         { return kDEFAULTMATERIALS;         }
  else if (temp == "SURFACEROUGHNESS")         { return kSURFACEROUGHNESS;         }
  else if (temp == "EMISSIVE")                 { return kEMISSIVE;                 }
  else if (temp == "AMBIENT")                  { return kAMBIENT;                  }
  else if (temp == "DIFFUSE")                  { return kDIFFUSE;                  }
  else if (temp == "SPECULAR")                 { return kSPECULAR;                 }
  else if (temp == "SPECULAREXPONENT")         { return kSPECULAREXPONENT;         }
  else if (temp == "LIGHTBACKFACES")           { return kLIGHTBACKFACES;           }
  else if (temp == "SPEEDOFLIGHT")             { return kSPEEDOFLIGHT;             }
  else if (temp == "TRANSPARENCY")             { return kTRANSPARENCY;             }
  else if (temp == "REFLECTIVITY")             { return kREFLECTIVITY;             }
  else if (temp == "GLOSSSTRENGTH")            { return kGLOSSSTRENGTH;            }
  else if (temp == "TRANSLUCENCYSTRENGTH")     { return kTRANSLUCENCYSTRENGTH;     }
  else if (temp == "PRIORITY")                 { return kPRIORITY;                 }
  else if (temp == "RETAINAMBIENTCOLOR")       { return kRETAINAMBIENTCOLOR;       }
  else if (temp == "RETAINDIFFUSECOLOR")       { return kRETAINDIFFUSECOLOR;       }
  else if (temp == "RETAINSPECULARREFLECTION") { return kRETAINSPECULARREFLECTION; }
  else if (temp == "DISABLEREFRACTION")        { return kDISABLEREFRACTION;        }
  else if (temp == "EYEPOINT")                 { return kEYEPOINT;                 }
  else if (temp == "VIEWPOINT")                { return kVIEWPOINT;                }
  else if (temp == "VIEWANGLE")                { return kVIEWANGLE;                }
  else if (temp == "UPDIRECTION")              { return kUPDIRECTION;              }
  else if (temp == "BACKGROUND")               { return kBACKGROUND;               }
  else if (temp == "GLOBALAMBIENT")            { return kGLOBALAMBIENT;            }
  else if (temp == "LIGHTATTENUATION")         { return kLIGHTATTENUATION;         }
  else if (temp == "ATMOSPHERICATTENUATION")   { return kATMOSPHERICATTENUATION;   }
  else if (temp == "FOG")                      { return kFOG;                      }
  else if (temp == "SHADOWFEELEREPSILON")      { return kSHADOWFEELEREPSILON;      }
  else if (temp == "REFLECTIVITYEPSILON")      { return kREFLECTIVITYEPSILON;      }
  else if (temp == "TRANSPARENCYEPSILON")      { return kTRANSPARENCYEPSILON;      }
  else if (temp == "MESHEPSILON")              { return kMESHEPSILON;              }
  else if (temp == "ADAPTEDSHADOWCOMPUTATION") { return kADAPTEDSHADOWCOMPUTATION; }
  else if (temp == "MINREFLECTIVITY")          { return kMINREFLECTIVITY;          }
  else if (temp == "MINTRANSPARENCY")          { return kMINTRANSPARENCY;          }
  else if (temp == "MAXRECURSIONDEPTH")        { return kMAXRECURSIONDEPTH;        }
  else if (temp == "SHOWEXTENTS")              { return kSHOWEXTENTS;              }
  else if (temp == "SOLIDTEXTURE")             { return kSOLIDTEXTURE;             }
  else if (temp == "IMAGETEXTURE")             { return kIMAGETEXTURE;             }
  else if (temp == "PROCEDURALTEXTURE")        { return kPROCEDURALTEXTURE;        }
  else if (temp == "FLATTEXTUREOFFSET")        { return kFLATTEXTUREOFFSET;        }
  else if (temp == "{")                        { return kLEFTBRACE;                }
  else if (temp == "}")                        { return kRIGHTBRACE;               }
  else if (temp == "DEF")                      { return kDEF;                      }
  else if (temp == "USE")                      { return kUSE;                      }
  else if (temp == " ")                        { return kT_NULL;                   }
  else if (isIdentifier(temp))                 { return kIDENT;                    }
  else                                         { return kUNKNOWN;                  }
}

double Scene::calcShadowFactor(const Ray& shadowFeeler)
{
  // normalization is done here because we need the real distance
  // to the light to compute light-source attenuation
  Ray normalizedShadowFeeler(shadowFeeler);
  normalizedShadowFeeler.fRecurseLevel = 0;

  if (fAdaptedShadowComputation) {
    normalizedShadowFeeler.fDirection.normalize();
  }

  double shadowFactor = 1.0;

  if (fComputeShadows) {

    double distanceToLight = shadowFeeler.fDirection.length();
    if (fAdaptedShadowComputation) {
      distanceToLight = 1.0;
    }

    if (fComputeSoftShadows) {

      // update statistics
      fNrOfRaysShot += normalizedShadowFeeler.fNrOfConeSamples;

      vector<Ray> jitteredShadowFeelers;
      Ray::jitterRaysInCone(jitteredShadowFeelers,
                            normalizedShadowFeeler,
                            fShadowFeelerJitterStrength / 2.0);

      double averageShadowFactor = 0.0;

      for (int coneSample = 0; coneSample < normalizedShadowFeeler.fNrOfConeSamples; ++coneSample) {

        // determine the amount of light that passes through
        // all objects (opaque and transparent) that the ray intersects

        GeometricObjects::const_iterator p;
        Intersection dummyIntersection;

        shadowFactor = 1.0;

        for (p = fGeometricObjects->begin(); p != fGeometricObjects->end(); ++p) {

          // note that non-convex objects can shadow themselves
          if ((*p)->computeIntersectionsWithRay(jitteredShadowFeelers[coneSample],dummyIntersection)) {

            if ((dummyIntersection.fHits[0].fHitTime > 0.0) &&
                (dummyIntersection.fHits[0].fHitTime < distanceToLight)) {
              shadowFactor *= dummyIntersection.fHits[0].fObjectHit->fMaterial.fTransparency;
            }
          }
        }
        
        averageShadowFactor += shadowFactor;
      }

      shadowFactor = averageShadowFactor / normalizedShadowFeeler.fNrOfConeSamples;
    }
    else { // compute hard shadows

      // update statistics
      ++fNrOfRaysShot;

      // determine the amount of light that passes through
      // all objects (opaque and transparent) that the ray intersects

      GeometricObjects::const_iterator p;
      Intersection dummyIntersection;

      for (p = fGeometricObjects->begin(); p != fGeometricObjects->end(); ++p) {

        // note that non-convex objects can shadow themselves
        if ((*p)->computeIntersectionsWithRay(normalizedShadowFeeler,dummyIntersection)) {
          if ((dummyIntersection.fHits[0].fHitTime > 0.0) &&
              (dummyIntersection.fHits[0].fHitTime < distanceToLight)) {
            shadowFactor *= dummyIntersection.fHits[0].fObjectHit->fMaterial.fTransparency;
          }
        }
      }
    }
  }

  return shadowFactor;
}

void Scene::resetMotionBlurTransformations(void)
{
  fCurrentMaterial.fMotionBlurEnabled = false;
  fMotionBlurRotationAngle = 0.0;
  fMotionBlurRotationAxis = Vector3D(0.0,1.0,0.0);
  fMotionBlurTranslation = Vector3D(0.0,0.0,0.0);
  fMotionBlurScale = Vector3D(1.0,1.0,1.0);
}
