// -------------------------------
// Filename      : solidtexturer.h
// Author        : Sven Maerivoet
// Last modified : 31/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// -------------------------------

#ifndef SM_SOLIDTEXTURER_H // include-guard
#define SM_SOLIDTEXTURER_H

#include "primitives.h"

class SolidTexturer {
  public:
    // some useful constants
    enum ESolidTextureType {kNone,
                            kCheckerboard,
                            kStackedColorBars,
                            kRGBCubes,
                            kOneCoordinateBasedContourLines,
                            kTwoCoordinatesBasedContourLines,
                            kThreeCoordinatesBasedContourLines,
                            kWoodGrain,
                            kMarble,
                            kCowSpots};

    // constructor
    SolidTexturer(void);

    // destructor
    ~SolidTexturer(void);

    // methods
    void initialize(void);
    Color sampleCheckerboard(const Point3D& p,
                             const Vector3D& scale) const;
    Color sampleStackedColorBars(const Point3D& p,
                                 const Vector3D& scale) const;
    Color sampleRGBCubes(const Point3D& p,
                         const Vector3D& scale) const;
    Color sampleOneCoordinateBasedContourLines(const Point3D& p,
                                               int coordinateInvolved,
                                               double scale) const;
    Color sampleTwoCoordinateBasedContourLines(const Point3D& p,
                                               int coordinatesInvolved,
                                               int distanceMeasureToUse,
                                               const Vector2D& scale) const;
    Color sampleThreeCoordinateBasedContourLines(const Point3D& p,
                                                 int distanceMeasureToUse,
                                                 const Vector3D& scale) const;
    Color sampleWoodGrain(const Point3D& p,
                          double ringThickness,
                          double wobbleStrength,
                          int nrOfWobbles,
                          double twistStrength,
                          const Color& color1,
                          const Color& color2,
                          bool interpolateSmoothly) const;
    Color sampleMarble(const Point3D& p,
                       int coordinateInvolved,
                       double turbulenceStrength,
                       double redWeight,
                       double greenWeight,
                       double blueWeight) const;
    Color sampleCowSpots(const Point3D& p,
                         double treshold) const;

  private:
    // data-members
    const double   kOffset;
    double*        fNoiseTable;
    unsigned char* fIndex;

    // utility methods
    double spline(double x) const;
    double turbulence(double scale, const Point3D& p) const;
    double noise(double scale, const Point3D& p) const;
    double latticeNoise(int i, int j, int k) const;
};

#endif
